from zeta.nn.masks.attn_masks import (
    AttentionBias,
    _materialize_causal_mask,
    LocalAttentionFromBottomRightMask,
    LowerTriangularMask,
    LowerTriangularFromBottomRightMask,
    LowerTriangularFromBottomRightLocalAttentionMask,
    LowerTriangularMaskWithTensorBias,
    _SeqLenInfo,
    _PaddedSeqLenInfo,
    BlockDiagonalMask,
    BlockDiagonalCausalMask,
    BlockDiagonalCausalFromBottomRightMask,
    BlockDiagonalCausalWithOffsetPaddedKeysMask,
    BlockDiagonalCausalLocalAttentionMask,
    BlockDiagonalCausalLocalAttentionFromBottomRightMask,
)

__all__ = [
    "AttentionBias",
    "_materialize_causal_mask",
    "LocalAttentionFromBottomRightMask",
    "LowerTriangularMask",
    "LowerTriangularFromBottomRightMask",
    "LowerTriangularFromBottomRightLocalAttentionMask",
    "LowerTriangularMaskWithTensorBias",
    "_SeqLenInfo",
    "_PaddedSeqLenInfo",
    "BlockDiagonalMask",
    "BlockDiagonalCausalMask",
    "BlockDiagonalCausalFromBottomRightMask",
    "BlockDiagonalCausalWithOffsetPaddedKeysMask",
    "BlockDiagonalCausalLocalAttentionMask",
    "BlockDiagonalCausalLocalAttentionFromBottomRightMask",
]



import torch
from torch import nn
import torch.nn.functional as F


def pad_at_dim(t, pad, dim=-1, value=0.0):
    dims_from_right = (-dim - 1) if dim < 0 else (t.ndim - dim - 1)
    zeros = (0, 0) * dims_from_right
    return F.pad(t, (*zeros, *pad), value=value)


def exists(val):
    return val is not None


def shift(t, amount, mask=None):
    if amount == 0:
        return t
    else:
        amount = min(amount, t.shape[-1])

    if exists(mask):
        t = t.masked_fill(~mask[..., None], 0.0)

    return pad_at_dim(t, (amount, -amount), dim=-2, value=0.0)


class ShiftTokens(nn.Module):
    """
    Shift Tokens

    Overview: Shift tokens in the input sequence

    Args:
        shifts (list): list of shifts
        fn (nn.Module): function to apply after shifting

    Attributes:
        shifts (tuple): tuple of shifts
        fn (nn.Module): function to apply after shifting

    Usage:
        >>> x = torch.randn(1, 10, 512)
        >>> shift_tokens = ShiftTokens(shifts=[1, 2, 3], fn=nn.Linear(512, 512))
        >>> shift_tokens(x).shape
        torch.Size([1, 10, 512])

    """

    def __init__(self, shifts, fn):
        super().__init__()
        self.fn = fn
        self.shifts = tuple(shifts)

    def forward(self, x, **kwargs):
        """Forward method of ShiftTokens"""
        mask = kwargs.get("mask", None)
        shifts = self.shifts
        segments = len(shifts)

        feats_per_shift = x.shape[-1] // segments
        splitted = x.split(feats_per_shift, dim=-1)
        segments_to_shift, rest = splitted[:segments], splitted[segments:]
        segments_to_shift = list(
            map(
                lambda args: shift(*args, mask=mask),
                zip(segments_to_shift, shifts),
            )
        )
        x = torch.cat((*segments_to_shift, *rest), dim=-1)
        return self.fn(x, **kwargs)



import copy
import torch
from torch import nn, Tensor
from zeta.utils.enforce_types import enforce_types


class SLERPModelMerger(nn.Module):
    """
    A class to merge models using Spherical Linear Interpolation (SLERP).

    SLERP provides a method to interpolate between two sets of weights, which can be
    beneficial for combining models trained in different phases.

    Attributes:
        model1 (nn.Module): The first model to be merged.
        model2 (nn.Module): The second model to be merged.
        t (float): The interpolation parameter ranging from 0 (model1) to 1 (model2).

    Examples::
    model1 = nn.Linear(10, 10)
    model2 = nn.Linear(10, 10)
    model3 = nn.Linear(10, 10)
    model4 = nn.Linear(10, 10)

    merge = SLERPModelMerger(model1, model2, 0.5)
    merged_model = merge.merge()
    print(merged_model.state_dict())
    """

    @enforce_types
    def __init__(
        self,
        model1: nn.Module,
        model2: nn.Module,
        t: float = 0.5,
    ):
        super().__init__()
        self.model1 = model1
        self.model2 = model2
        self.t = t

    def merge(self) -> nn.Module:
        """
        Merges the models using SLERP.

        Returns:
            nn.Module: A new model with merged weights.
        """
        merged_model = self._copy_model_structure(self.model1)

        # Get the state dicts of both models
        state_dict1 = self.model1.state_dict()
        state_dict2 = self.model2.state_dict()

        # Init a state dict for the merged model
        merged_state_dict = merged_model.state_dict()

        for key in merged_state_dict.keys():
            # Perform WELP for each parameter
            w1 = state_dict1[key]
            w2 = state_dict2[key]
            merged_state_dict[key] = self._slerp(w1, w2, self.t)

        # Load the mergd state dict into the new model
        merged_model.load_state_dict(merged_state_dict)
        return merged_model

    @staticmethod
    @enforce_types
    def _slerp(w1: Tensor, w2: Tensor, t: float) -> Tensor:
        """
        Performs Spherical Linear Interpolation (SLERP) between two tensors.

        Args:
            w1 (torch.Tensor): The first tensor.
            w2 (torch.Tensor): The second tensor.
            t (float): The interpolation parameter.

        Returns:
            torch.Tensor: The interpolated tensor.
        """
        omega = torch.acos(
            torch.clamp(
                torch.dot(w1.view(-1), w2.view(-1))
                / (torch.norm(w1) * torch.norm(w2)),
                -1,
                1,
            )
        )
        sin_omega = torch.sin(omega)
        return (torch.sin((1.0 - t) * omega) / sin_omega) * w1 + (
            torch.sin(t * omega) / sin_omega
        ) * w2

    @staticmethod
    @enforce_types
    def _copy_model_structure(model: nn.Module) -> nn.Module:
        """
        Creates a new instance of a model with the same structure as the given model.

        Args:
            model (nn.Module): The model whose structure is to be copied.

        Returns:
            nn.Module: A new model with the same structure.
        """
        assert isinstance(
            model, nn.Module
        ), "model must be an nn.Module instance"
        model_copy = copy.deepcopy(model)
        return model_copy


# model1 = nn.Linear(10, 10)
# model2 = nn.Linear(10, 10)
# model3 = nn.Linear(10, 10)
# model4 = nn.Linear(10, 10)

# merge = SLERPModelMerger(model1, model2, 0.5)
# merged_model = merge.merge()
# print(merged_model.state_dict())



import torch
from torch import nn, Tensor
from zeta.nn.modules.mbconv import MBConv
from zeta.nn.modules.film import Film


class FiLMEfficientNetB3(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        dim: int,
        downsample: int,
        kernel_size: int,
        stride: int,
        padding: int,
        dropout: float = 0.1,
        num_mbconv_blocks: int = 26,
        num_film_layers: int = 26,
        expanse_ratio: int = 4,
        *args,
        **kwargs,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.dim = dim
        self.num_mbconv_blocks = num_mbconv_blocks
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.num_film_layers = num_film_layers
        self.expanse_ratio = expanse_ratio
        self.hidden_dim = dim * expanse_ratio

        for _ in range(num_mbconv_blocks):
            self.mb_conv_layers = nn.ModuleList(
                [
                    MBConv(
                        dim_in=in_channels,
                        dim_out=dim,
                        downsample=downsample,
                        dropout=dropout,
                        *args,
                        **kwargs,
                    )
                ]
            )

            self.film_layers = nn.ModuleList(
                [Film(dim, self.hidden_dim, expanse_ratio=expanse_ratio)]
            )

        self.proj = nn.Linear(in_channels, out_channels)

    def forward(
        self, text: Tensor, img: Tensor, weight: Tensor = None, *args, **kwargs
    ) -> Tensor:
        x = img

        # Apply MBConv and film layers
        for mb_conv, film in zip(self.mb_conv_layers, self.film_layers):
            x = mb_conv(x)
            x = film(x, text)

        # Flatten the output to pass through the projection layer
        x = x.view(x.size(0), -1)
        x = self.proj(x)

        return x


# Assuming the MBConv and Film layers are properly defined in the modules,
# the FiLMEfficientNetB3 can be instantiated and used as follows:

# Example usage
film_efficient_net = FiLMEfficientNetB3(
    in_channels=512,
    out_channels=1000,
    dim=512,
    downsample=1,
    kernel_size=3,
    stride=1,
    padding=1,
    dropout=0.1,
)

# Mock inputs
text_input = torch.randn(1, 512)  # Example text input
img_input = torch.randn(1, 3, 224, 224)  # Example image input

# Forward pass
output = film_efficient_net(text_input, img_input)
print(
    output.shape
)  # Expected shape: (1, 1000), which depends on the final projection layer



# Copyright (c) 2022 Agora
# Licensed under The MIT License [see LICENSE for details]
from zeta.utils.cuda_memory_wrapper import track_cuda_memory_usage

from zeta.utils.benchmark import (
    benchmark,
    print_cuda_memory_usage,
    save_memory_snapshot,
)
from zeta.utils.disable_logging import disable_warnings_and_logs
from zeta.utils.params import print_num_params, print_main
from zeta.utils.module_device import module_device
from zeta.utils.save_load_wrapper import save_load
from zeta.utils.main import (
    exists,
    default,
    once,
    eval_decorator,
    cast_tuple,
    maybe,
    init_zero_,
    pick_and_pop,
    group_dict_by_key,
    string_begins_with,
    group_by_key_prefix,
    top_p,
    top_k,
    top_a,
    log,
    gumbel_noise,
    video_tensor_to_gift,
    gif_to_tensor,
    l2norm,
    pad_at_dim,
    cosine_beta_schedule,
    cast_if_src_dtype,
    get_sinusoid_encoding_table,
    interpolate_pos_encoding_2d,
)

from zeta.utils.enforce_types import enforce_types

#######
from zeta.utils.cuda_wrapper import (
    get_cuda_bare_metal_version,
    check_cuda_torch_binary_vs_bare_metal,
    raise_if_cuda_home_none,
    append_nvcc_threads,
    check_cuda,
)
from zeta.utils.verbose_execution import VerboseExecution


####
__all__ = [
    "track_cuda_memory_usage",
    "benchmark",
    "print_cuda_memory_usage",
    "save_memory_snapshot",
    "disable_warnings_and_logs",
    "print_main",
    "module_device",
    "save_load",
    "exists",
    "default",
    "once",
    "eval_decorator",
    "cast_tuple",
    "maybe",
    "init_zero_",
    "pick_and_pop",
    "group_dict_by_key",
    "string_begins_with",
    "group_by_key_prefix",
    "top_p",
    "top_k",
    "top_a",
    "log",
    "gumbel_noise",
    "print_num_params",
    "video_tensor_to_gift",
    "gif_to_tensor",
    "l2norm",
    "pad_at_dim",
    "cosine_beta_schedule",
    "cast_if_src_dtype",
    "get_sinusoid_encoding_table",
    "interpolate_pos_encoding_2d",
    "enforce_types",
    "get_cuda_bare_metal_version",
    "check_cuda_torch_binary_vs_bare_metal",
    "raise_if_cuda_home_none",
    "append_nvcc_threads",
    "check_cuda",
    "VerboseExecution",
]



from dataclasses import dataclass

import torch
from einops import rearrange
from torch import Tensor, nn

from zeta.nn.modules.image_to_text import img_to_text


@dataclass
class GILLMapper(nn.Module):
    """
    GILLMapper is a module that maps image and text embeddings using a Transformer model.
    From the paper: "https://arxiv.org/pdf/2305.17216.pdf"

    Args:
        img_emb_size (int): The size of the image embeddings.
        text_emb_size (int): The size of the text embeddings.
        num_encoder_layers (int): The number of layers in the encoder of the Transformer model.
        num_decoder_layers (int): The number of layers in the decoder of the Transformer model.
        heads (int): The number of attention heads in the Transformer model.
        dim_ffn (int): The size of the feed-forward neural network in the Transformer model.
        seq_length (int): The length of the input sequence.
        dropout (float, optional): The dropout rate. Defaults to 0.1.
        args (dict, optional): Additional arguments. Defaults to None.

    Example:
        >>> model = GILLMapper(
        ...     img_emb_size=512,
        ...     text_emb_size=512,
        ...     num_encoder_layers=6,
        ...     num_decoder_layers=6,
        ...     heads=8,
        ...     dim_ffn=2048,
        ...     seq_length=100,
        ...     dropout=0.1,
        ...     args=None
        ... )
        >>> img = torch.randn(1, 3, 224, 224)
        >>> text = torch.randn(1, 100, 512)
        >>> out = model(img, text)
        >>> out.shape
    """

    img_emb_size: int
    text_emb_size: int
    num_encoder_layers: int
    num_decoder_layers: int
    heads: int
    dim_ffn: int
    seq_length: int
    dropout: float = 0.1
    args: dict = None

    def __post_init__(self):
        super(GILLMapper, self).__init__()
        self.transformer = nn.Transformer(
            d_model=self.text_emb_size,
            num_encoder_layers=self.num_encoder_layers,
            num_decoder_layers=self.num_decoder_layers,
            dim_feedforward=self.dim_ffn,
        )
        self.img_to_text_proj = nn.Linear(self.img_emb_size, self.text_emb_size)
        self.learned_queries = nn.Parameter(
            torch.randn(self.seq_length, self.text_emb_size)
        )
        self.output_layer = nn.Linear(self.text_emb_size, self.text_emb_size)
        self.text_embedding_layer = nn.Embedding(
            self.seq_length, self.text_emb_size
        )
        self.img_embedding_layer = nn.Linear(
            self.img_emb_size, self.text_emb_size
        )

        self.transformer_encoder = nn.TransformerEncoder(
            nn.TransformerEncoderLayer(
                d_model=self.text_emb_size,
                nhead=self.heads,
                dim_feedforward=self.dim_ffn,
            ),
            num_layers=self.num_encoder_layers,
        )

    def forward(self, img: Tensor, text: Tensor) -> Tensor:
        """
        Forward pass of the GILLMapper module.

        Args:
            img (Tensor): The input image tensor. 4D tensor of shape (B, C, H, W).
            text (Tensor): The input text tensor. 3D tensor of shape (batch_size, seq_length).

        Returns:
            Tensor: The output tensor.
        """
        # Embed the image and text
        # img = self.img_embedding_layer(img)
        text = self.text_embedding_layer(text)

        t_b, t_n, t_d = text.shape
        img = img_to_text(img, t_n, t_d)

        # Transforming the img with the encoder
        img = self.transformer_encoder(img)
        print(f"img shape: {img.shape}")

        # Rearrange embeddings for transformer
        img = rearrange(img, "b n d -> n b d ")
        text = rearrange(text, "b n d -> n b d ")

        # Expand learned queries to match the batch
        queries = rearrange(self.learned_queries, "n d -> n 1 d").expand(
            -1, img.shape[1], -1
        )

        # Transformer
        output = self.transformer(src=img, tgt=queries + text)

        # Output layer
        out = self.output_layer(output)
        out = rearrange(out, "n b d -> b n d")

        return out



import torch
import torch.nn as nn
from einops import repeat


class TextSceneAttentionFusion(nn.Module):
    """
    TextSceneAttentionFusion is an attention-based fusion mechanism to combine
    text sequences and 3D scene embeddings. The model computes attention scores
    for each text token with respect to the 3D scene to focus on specific parts
    of the text that are more relevant to the scene.

    Args:
        text_features (int): Dimension of the text embeddings.
        scene_features (int): Dimension of the scene embeddings.

    Forward Args:
        text (torch.Tensor): Tensor of shape [batch_size, seq_len, text_features]
                             representing sequences of text embeddings.
        scene (torch.Tensor): Tensor of shape [batch_size, depth, height, width, scene_features]
                              representing 3D scene embeddings.

    Returns:
        torch.Tensor: Tensor of shape [batch_size, seq_len, scene_features]
                      representing the fused embeddings.
    """

    def __init__(self, text_features: int, scene_features: int):
        super(TextSceneAttentionFusion, self).__init__()

        # A linear layer for calculating attention scores
        self.attention = nn.Linear(text_features + scene_features, 1)

    def forward(self, text: torch.Tensor, scene: torch.Tensor) -> torch.Tensor:
        # Flattening spatial dimensions of the scene for simplicity
        batch_size, depth, height, width, scene_features = scene.shape
        scene_flat = scene.view(
            batch_size, depth * height * width, scene_features
        )

        # Using einops to repeat the scene tensor for matching text sequence length
        scene_expanded = repeat(
            scene_flat, "b sh sf -> b st sh sf", st=text.size(1)
        )

        # Repeating the text tensor to match the flattened spatial dimensions of the scene
        text_expanded = repeat(
            text, "b st tf -> b st sh tf", sh=depth * height * width
        )

        # Concatenating expanded scene tensor and text tensor
        concat_features = torch.cat(
            [text_expanded, scene_expanded], dim=-1
        )  # [batch_size, seq_len, depth*height*width, text_features + scene_features]

        # Computing attention scores and weights
        attention_scores = self.attention(concat_features).squeeze(
            -1
        )  # [batch_size, seq_len, depth*height*width]
        attention_weights = torch.nn.functional.softmax(
            attention_scores.flatten(2), dim=-1
        ).view(batch_size, seq_len, depth * height * width, 1)

        # Using einsum to obtain weighted scene embeddings
        fused = torch.einsum(
            "btsh,btshj->btsj", attention_weights, scene_expanded
        )

        return fused


# Usage Test
if __name__ == "__main__":
    text_features = 512
    scene_features = 1024
    model = TextSceneAttentionFusion(text_features, scene_features)

    batch_size = 16
    seq_len = 10
    depth = 4
    height = 64
    width = 64

    text = torch.rand(batch_size, seq_len, text_features)
    scene = torch.rand(batch_size, depth, height, width, scene_features)

    fused = model(text, scene)
    print(
        f"Fused output shape: {fused.shape}"
    )  # Expected: [batch_size, seq_len, scene_features]



"""
Lookup Free Quantization
Proposed in https://arxiv.org/abs/2310.05737

In the simplest setup, each dimension is quantized into {-1, 1}.
An entropy penalty is used to encourage utilization.
"""

from collections import namedtuple
from math import ceil, log2

import torch
import torch.nn.functional as F
from einops import pack, rearrange, reduce, unpack
from torch import Tensor, einsum, nn
from torch.nn import Module

# constants

Return = namedtuple("Return", ["quantized", "indices", "entropy_aux_loss"])

LossBreakdown = namedtuple(
    "LossBreakdown", ["per_sample_entropy", "batch_entropy", "commitment"]
)

# helper functions


def exists(v):
    return v is not None


def default(*args):
    for arg in args:
        if exists(arg):
            return arg() if callable(arg) else arg
    return None


def pack_one(t, pattern):
    return pack([t], pattern)


def unpack_one(t, ps, pattern):
    return unpack(t, ps, pattern)[0]


# entropy


def log(t, eps=1e-5):
    return t.clamp(min=eps).log()


def entropy(prob):
    return (-prob * log(prob)).sum(dim=-1)


# class


class LFQ(Module):
    """
    Initializes the Lookup-Free Quantization (LFQ) module.

    Args:
        dim (int, optional): The input dimension. If not specified, it is calculated based on the codebook size and number of codebooks. Defaults to None.
        codebook_size (int, optional): The size of the codebook. If not specified, it is calculated based on the input dimension. Defaults to None.
        entropy_loss_weight (float, optional): The weight for the entropy loss. Defaults to 0.1.
        commitment_loss_weight (float, optional): The weight for the commitment loss. Defaults to 0.25.
        diversity_gamma (float, optional): The gamma parameter for diversity regularization. Defaults to 1.0.
        straight_through_activation (nn.Module, optional): The activation function to be used during the forward pass. Defaults to nn.Identity().
        num_codebooks (int, optional): The number of codebooks. Defaults to 1.
        keep_num_codebooks_dim (bool, optional): Whether to keep the number of codebooks dimension. Defaults to None.
        codebook_scale (float, optional): The scale factor for the codebook. Defaults to 1.0.

    Examples::
        import torch
        from zeta.nn import LFQ

        # you can specify either dim or codebook_size
        # if both specified, will be validated against each other

        quantizer = LFQ(
            codebook_size = 65536,      # codebook size, must be a power of 2
            dim = 16,                   # this is the input feature dimension, defaults to log2(codebook_size) if not defined
            entropy_loss_weight = 0.1,  # how much weight to place on entropy loss
            diversity_gamma = 1.        # within entropy loss, how much weight to give to diversity of codes, taken from https://arxiv.org/abs/1911.05894
        )

        image_feats = torch.randn(1, 16, 32, 32)

        quantized, indices, entropy_aux_loss = quantizer(image_feats)

        # (1, 16, 32, 32), (1, 32, 32), (1,)

        assert image_feats.shape == quantized.shape
        assert (quantized == quantizer.indices_to_codes(indices)).all()
    """

    def __init__(
        self,
        *,
        dim=None,
        codebook_size=None,
        entropy_loss_weight=0.1,
        commitment_loss_weight=0.25,
        diversity_gamma=1.0,
        straight_through_activation=nn.Identity(),
        num_codebooks=1,
        keep_num_codebooks_dim=None,
        codebook_scale=1.0,  # for residual LFQ, codebook scaled down by 2x at each layer
    ):
        super().__init__()

        # some assert validations

        assert exists(dim) or exists(
            codebook_size
        ), "either dim or codebook_size must be specified for LFQ"
        assert not exists(codebook_size) or log2(codebook_size).is_integer(), (
            "your codebook size must be a power of 2 for lookup free"
            f" quantization (suggested {2 ** ceil(log2(codebook_size))})"
        )

        codebook_size = default(codebook_size, lambda: 2**dim)
        codebook_dim = int(log2(codebook_size))

        codebook_dims = codebook_dim * num_codebooks
        dim = default(dim, codebook_dims)

        has_projections = dim != codebook_dims
        self.project_in = (
            nn.Linear(dim, codebook_dims) if has_projections else nn.Identity()
        )
        self.project_out = (
            nn.Linear(codebook_dims, dim) if has_projections else nn.Identity()
        )
        self.has_projections = has_projections

        self.dim = dim
        self.codebook_dim = codebook_dim
        self.num_codebooks = num_codebooks

        keep_num_codebooks_dim = default(
            keep_num_codebooks_dim, num_codebooks > 1
        )
        assert not (num_codebooks > 1 and not keep_num_codebooks_dim)
        self.keep_num_codebooks_dim = keep_num_codebooks_dim

        # straight through activation

        self.activation = straight_through_activation

        # entropy aux loss related weights

        self.diversity_gamma = diversity_gamma
        self.entropy_loss_weight = entropy_loss_weight

        # codebook scale

        self.codebook_scale = codebook_scale

        # commitment loss

        self.commitment_loss_weight = commitment_loss_weight

        # for no auxiliary loss, during inference

        self.register_buffer(
            "mask", 2 ** torch.arange(codebook_dim - 1, -1, -1)
        )
        self.register_buffer("zero", torch.tensor(0.0), persistent=False)

        # codes

        all_codes = torch.arange(codebook_size)
        bits = ((all_codes[..., None].int() & self.mask) != 0).float()
        codebook = self.bits_to_codes(bits)

        self.register_buffer("codebook", codebook, persistent=False)

    def bits_to_codes(self, bits):
        return bits * self.codebook_scale * 2 - self.codebook_scale

    @property
    def dtype(self):
        return self.codebook.dtype

    def indices_to_codes(self, indices, project_out=True):
        """Indices to codes.

        Args:
            indices (_type_): _description_
            project_out (bool, optional): _description_. Defaults to True.

        Returns:
            _type_: _description_
        """
        is_img_or_video = indices.ndim >= (3 + int(self.keep_num_codebooks_dim))

        if not self.keep_num_codebooks_dim:
            indices = rearrange(indices, "... -> ... 1")

        # indices to codes, which are bits of either -1 or 1

        bits = ((indices[..., None].int() & self.mask) != 0).to(self.dtype)

        codes = self.bits_to_codes(bits)

        codes = rearrange(codes, "... c d -> ... (c d)")

        # whether to project codes out to original dimensions
        # if the input feature dimensions were not log2(codebook size)

        if project_out:
            codes = self.project_out(codes)

        # rearrange codes back to original shape

        if is_img_or_video:
            codes = rearrange(codes, "b ... d -> b d ...")

        return codes

    def forward(
        self,
        x: Tensor,
        inv_temperature=100.0,
        return_loss_breakdown=False,
        mask=None,
    ) -> Tensor:
        """
        einstein notation
        b - batch
        n - sequence (or flattened spatial dimensions)
        d - feature dimension, which is also log2(codebook size)
        c - number of codebook dim
        """

        is_img_or_video = x.ndim >= 4

        # standardize image or video into (batch, seq, dimension)

        if is_img_or_video:
            x = rearrange(x, "b d ... -> b ... d")
            x, ps = pack_one(x, "b * d")

        assert (
            x.shape[-1] == self.dim
        ), f"expected dimension of {self.dim} but received {x.shape[-1]}"

        x = self.project_in(x)

        # split out number of codebooks

        x = rearrange(x, "b n (c d) -> b n c d", c=self.num_codebooks)

        # quantize by eq 3.

        original_input = x

        codebook_value = torch.ones_like(x) * self.codebook_scale
        quantized = torch.where(x > 0, codebook_value, -codebook_value)

        # use straight-through gradients (optionally with custom activation fn) if training

        if self.training:
            x = self.activation(x)
            x = x + (quantized - x).detach()
        else:
            x = quantized

        # calculate indices

        indices = reduce(
            (x > 0).int() * self.mask.int(), "b n c d -> b n c", "sum"
        )

        # entropy aux loss

        if self.training:
            # the same as euclidean distance up to a constant
            distance = -2 * einsum(
                "... i d, j d -> ... i j", original_input, self.codebook
            )

            prob = (-distance * inv_temperature).softmax(dim=-1)

            per_sample_entropy = entropy(prob).mean()

            # account for mask

            if exists(mask):
                prob = prob[mask]

            # distribution over all available tokens in the batch

            avg_prob = reduce(prob, "... c d -> c d", "mean")
            codebook_entropy = entropy(avg_prob).mean()

            # 1. entropy will be nudged to be low for each code, to encourage the network to output confident predictions
            # 2. codebook entropy will be nudged to be high, to encourage all codes to be uniformly used within the batch

            entropy_aux_loss = (
                per_sample_entropy - self.diversity_gamma * codebook_entropy
            )
        else:
            # if not training, just return dummy 0
            entropy_aux_loss = per_sample_entropy = codebook_entropy = self.zero

        # commit loss

        if self.training:
            commit_loss = F.mse_loss(
                original_input, quantized.detach(), reduction="none"
            )

            if exists(mask):
                commit_loss = commit_loss[mask]

            commit_loss = commit_loss.mean()
        else:
            commit_loss = self.zero

        # merge back codebook dim

        x = rearrange(x, "b n c d -> b n (c d)")

        # project out to feature dimension if needed

        x = self.project_out(x)

        # reconstitute image or video dimensions

        if is_img_or_video:
            x = unpack_one(x, ps, "b * d")
            x = rearrange(x, "b ... d -> b d ...")

            indices = unpack_one(indices, ps, "b * c")

        # whether to remove single codebook dim

        if not self.keep_num_codebooks_dim:
            indices = rearrange(indices, "... 1 -> ...")

        # complete aux loss

        aux_loss = (
            entropy_aux_loss * self.entropy_loss_weight
            + commit_loss * self.commitment_loss_weight
        )

        ret = Return(x, indices, aux_loss)

        if not return_loss_breakdown:
            return ret

        return ret, LossBreakdown(
            per_sample_entropy, codebook_entropy, commit_loss
        )



import logging
import math
import time
from typing import Any, Callable, Dict, Optional, Tuple, cast

import torch
import torch.distributed as dist
import torch.nn.functional as F
from torch import Tensor
from torch.nn import Module, ModuleList

try:
    from fairseq.modules.moe import MOELayer

    has_fairseq = True
    Base = MOELayer
except ModuleNotFoundError:
    Base = Module
    has_fairseq = False

try:
    # To enable Tutel MoE optimizations:
    #   python3 -m pip install --user --upgrade git+https://github.com/microsoft/tutel@v0.1.x
    from tutel import moe as tutel_moe

    has_tutel, fused_cumsum_sub_one = True, tutel_moe.fast_cumsum_sub_one
except ModuleNotFoundError:
    has_tutel, fused_cumsum_sub_one = (
        False,
        lambda mask: torch.cumsum(mask, dim=0) - 1,
    )

logger = logging.getLogger(__name__)


# use a fixed temperature to compute balance loss
TEMPERATURE_FOR_L_UAX = 0.07

# maximum capacity of 1 expert as a fraction of number of tokens in the batch
# Note: setting this to 1.0 causes inference to significantly slow down
EVAL_CAPACITY_TOKEN_FRACTION = 0.25

# logging
SAMPLE_FRACTION = 0.2


def _find_my_group_index(grouped_ranks):
    my_rank = dist.get_rank()
    for i, group in enumerate(grouped_ranks):
        if my_rank in group:
            return i
    raise RuntimeError


def get_moe_group(moe_expert_count=None):
    if dist.is_initialized():
        if not hasattr(get_moe_group, "_moe_groups"):
            world_size = dist.get_world_size()

            if world_size <= moe_expert_count:
                assert moe_expert_count % world_size == 0
                moe_groups = [[i] for i in range(world_size)]

            else:
                assert world_size % moe_expert_count == 0
                ranks_per_group = world_size // moe_expert_count
                moe_groups = [
                    [i + j * moe_expert_count for j in range(ranks_per_group)]
                    for i in range(moe_expert_count)
                ]

            get_moe_group._moe_expert_count = moe_expert_count
            get_moe_group._moe_group_idx = moe_groups
            get_moe_group._moe_groups = [dist.new_group(g) for g in moe_groups]

        my_group_idx = _find_my_group_index(get_moe_group._moe_group_idx)
        return my_group_idx, get_moe_group._moe_groups[my_group_idx]


def get_all2all_group(moe_expert_count):
    if dist.is_initialized():
        if not hasattr(get_all2all_group, "_all2all_groups"):
            world_size = dist.get_world_size()

            # more experts than world size
            if world_size <= moe_expert_count:
                assert moe_expert_count % world_size == 0
                all2all_groups = [[i for i in range(world_size)]]

            # larger world than num experts
            else:
                assert world_size % moe_expert_count == 0
                ranks_per_group = world_size // moe_expert_count
                all2all_groups = [
                    [i * moe_expert_count + j for j in range(moe_expert_count)]
                    for i in range(ranks_per_group)
                ]

            get_all2all_group._all2all_group_idx = all2all_groups
            get_all2all_group._all2all_groups = [
                dist.new_group(g) for g in all2all_groups
            ]

        my_group_idx = _find_my_group_index(
            get_all2all_group._all2all_group_idx
        )
        return get_all2all_group._all2all_groups[my_group_idx]


def top1gating(
    logits: torch.Tensor,
    input_mask: Optional[torch.Tensor] = None,
    use_fp32=False,
    capacity_factor=1.0,
    eval_mode=False,
    moe_eval_capacity_token_fraction=EVAL_CAPACITY_TOKEN_FRACTION,
    use_xmoe=False,
    gate_obj=None,
) -> Tuple[Tensor, Tensor, Tensor, Dict]:
    """Implements Top2Gating on logits."""
    metadata = {}
    if use_fp32:
        orig_dtype = logits.dtype
        logits = logits.float()

    gates = F.softmax(logits, dim=1)
    metadata["entropy_gating"] = entropy(probs=gates).mean().detach()

    # gates has shape of SE
    num_tokens = gates.shape[0]
    num_experts = gates.shape[1]
    if moe_eval_capacity_token_fraction > 0.0 and eval_mode:
        capacity = math.ceil(moe_eval_capacity_token_fraction * num_tokens)
    else:
        # capacity = capacity_factor * S/E
        capacity = int(capacity_factor * math.ceil(num_tokens / num_experts))

    # Create a mask for 1st's expert per token
    indices1_s = torch.argmax(gates, dim=1)
    mask1 = one_hot(indices1_s, num_classes=num_experts, unsqueeze_indices=True)
    if input_mask is not None and input_mask.any():
        nonpadding = ~input_mask
        mask1 = mask1 * nonpadding.unsqueeze(-1).to(mask1.dtype)

    # for logging (percent of tokens routed to each expert)
    expert1_hist = (
        100
        * torch.histc(
            (indices1_s.squeeze() + 1), bins=num_experts, min=1, max=num_experts
        )
        / num_tokens
    )
    metadata["unused_expert1_count"] = (expert1_hist == 0).sum()
    expert1_hist = (
        torch.sort(expert1_hist, dim=0, descending=True).values
        + torch.finfo(torch.float32).tiny
    )

    sample_count = max(math.ceil(num_experts * SAMPLE_FRACTION), 1)
    metadata["expert1_balance_top"] = expert1_hist[:sample_count].sum()
    metadata["expert1_balance_bottom"] = expert1_hist[-sample_count:].sum()

    gates1_s = (gates * mask1).sum(dim=1)

    # Compute locations in capacity buffer
    locations1 = fused_cumsum_sub_one(mask1)

    # Compute l_aux
    me = torch.mean(gates, dim=0)
    ce = torch.mean(mask1.to(gates.dtype), dim=0)

    l_aux = torch.mean(me * ce)
    l_aux = l_aux * num_experts * num_experts

    if has_tutel:
        locations1_s = torch.sum(locations1 * mask1, dim=1)
        return (
            l_aux,
            metadata,
            capacity,
            num_experts,
            [
                indices1_s,
            ],
            [
                locations1_s,
            ],
            [
                gates1_s,
            ],
        )

    # Remove locations outside capacity from mask
    mask1 = mask1 * torch.lt(locations1, capacity)
    # Store the capacity location for each token
    locations1_s = torch.sum(locations1 * mask1, dim=1)

    # Calculate combine_weights and dispatch_mask
    gates1 = gates1_s.unsqueeze(-1) * mask1.to(
        gates1_s.dtype
    )  # einsum("s,se->se")
    # locations1_sc = num_tokens * capacity
    locations1_sc = one_hot(
        locations1_s, num_classes=capacity, unsqueeze_indices=True
    )
    combine1_sec = torch.bmm(
        # einsum("se,sc->sec")
        gates1.unsqueeze(-1),
        locations1_sc.to(gates1.dtype).unsqueeze(1),
    )
    dispatch_mask = combine1_sec.bool()
    if use_fp32:
        return l_aux, combine1_sec.to(orig_dtype), dispatch_mask, metadata
    else:
        return l_aux, combine1_sec, dispatch_mask, metadata


class Top1Gate(torch.nn.Module):
    """Gate module which implements Top2Gating as described in Gshard_.
    ::

        gate = Top2Gate(model_dim, num_experts)
        l_aux, combine_weights, dispatch_mask = gate(input)

    .. Gshard_: https://arxiv.org/pdf/2006.16668.pdf

    Args:
        model_dim (int):
            size of model embedding dimension
        num_experts (ints):
            number of experts in model
    """

    wg: torch.nn.Linear

    def __init__(
        self,
        model_dim: int,
        num_experts: int,
        use_fp32=False,
        input_noise_type=None,
        capacity_factor=1.0,
        moe_eval_capacity_token_fraction=EVAL_CAPACITY_TOKEN_FRACTION,
        use_xmoe=False,
    ) -> None:
        # TODO: merge this to top2gate.py
        #
        super().__init__()

        if not use_xmoe:
            self.wg = torch.nn.Linear(model_dim, num_experts, bias=False)
        else:
            self.wg_reduction = torch.nn.Linear(model_dim, 16, bias=False)
            wg = torch.empty(num_experts, 16)
            torch.nn.init.orthogonal_(wg, gain=0.32)
            self.register_parameter("wg", torch.nn.Parameter(wg))

        self.use_xmoe = use_xmoe
        self.use_fp32 = use_fp32
        self.input_noise_type = input_noise_type
        self.capacity_factor = capacity_factor
        self.moe_eval_capacity_token_fraction = moe_eval_capacity_token_fraction

    def forward(self, input, mask=None):  # type: ignore
        if self.use_xmoe:
            input = self.wg_reduction(input)
            with torch.no_grad():
                wg_norm = self.wg.norm(p=2.0, dim=1, keepdim=True)
                self.wg.mul_(1.5 / wg_norm)
            logits = self._cosine(input, self.wg)
            logits = self._make_finite(logits)
        else:
            logits = self.wg(input)

        return top1gating(
            logits,
            mask,
            use_fp32=self.use_fp32,
            capacity_factor=self.capacity_factor,
            eval_mode=not self.training,
            moe_eval_capacity_token_fraction=self.moe_eval_capacity_token_fraction,
            use_xmoe=self.use_xmoe,
            gate_obj=self,
        )

    def _make_finite(self, scores):
        ok = scores.isfinite()
        if not ok.all():
            # NaNs here can break the assignment algorithm
            scores[~ok] = scores[ok].min()
        return scores

    def _get_gating_temperature(self, eps=1e-4):
        if self.gating_t.data.item() < eps:
            return eps
        return self.gating_t

    def _cosine(self, mat1, mat2, eps=1e-4):
        assert mat1.dim() == 2
        assert mat2.dim() == 2
        # mat1 = F.normalize(mat1, p=2.0, dim=1, eps=eps)
        mat2 = F.normalize(mat2.float(), p=2.0, dim=1, eps=eps)
        return mat1.float().matmul(mat2.transpose(0, 1)).type_as(mat1)


gumbel_map: Dict[torch.device, Callable] = {}


def gumbel_rsample(shape: Tuple, device: torch.device) -> Tensor:
    gumbel = gumbel_map.get(device)
    if gumbel is None:
        one = torch.tensor(1.0, device=device)
        zero = torch.tensor(0.0, device=device)
        gumbel = torch.distributions.gumbel.Gumbel(zero, one).rsample  # type: ignore
        gumbel_map[device] = gumbel
    return gumbel(shape)


def one_hot(
    indices: torch.Tensor, num_classes: int, unsqueeze_indices=False
) -> Tensor:
    if unsqueeze_indices:
        indices = indices.unsqueeze(-1)
    assert (
        indices.shape[-1] == 1
    ), "last dimension of indices must be have size 1"
    output = torch.zeros(
        indices.shape[:-1] + (num_classes,),
        device=indices.device,
        dtype=indices.dtype,
    )
    output.scatter_(len(output.shape) - 1, indices, 1)
    return output


def entropy(probs):
    logits = torch.distributions.utils.probs_to_logits(probs)
    p_log_p = probs * logits
    return -p_log_p.sum(-1)


def top2gating(
    logits: torch.Tensor,
    input_mask: Optional[torch.Tensor] = None,
    use_fp32=False,
    second_expert_policy="sampling",
    normalize_gate_prob_before_dropping=False,
    eval_mode=False,
    moe_eval_capacity_token_fraction=0.25,
    batch_prioritized_routing=False,
) -> Tuple[Tensor, Tensor, Tensor]:
    """Implements Top2Gating on logits."""
    metadata = {}
    if use_fp32:
        orig_dtype = logits.dtype
        logits = logits.float()
    gates = F.softmax(logits, dim=1)
    metadata["entropy_gating"] = entropy(probs=gates).mean().detach()
    # gates has shape of SE
    num_tokens = gates.shape[0]
    num_experts = gates.shape[1]
    if moe_eval_capacity_token_fraction > 0.0 and eval_mode:
        capacity = math.ceil(moe_eval_capacity_token_fraction * num_tokens)
    else:
        # capacity = 2S/E
        capacity = 2 * math.ceil(num_tokens / num_experts)

    # Create a mask for 1st's expert per token
    indices1_s = torch.argmax(gates, dim=1, keepdim=True)
    mask1 = one_hot(indices1_s, num_experts)
    if second_expert_policy == "sampling":
        # Create a mask for 2nd's expert per token using Gumbel-max trick
        # https://timvieira.github.io/blog/post/2014/07/31/gumbel-max-trick/
        logits_w_noise = logits + gumbel_rsample(
            logits.shape, device=logits.device
        )
    else:
        logits_w_noise = logits
    # Replace top-expert with min value
    logits_except1 = logits_w_noise.masked_fill(mask1.bool(), float("-inf"))
    indices2_s = torch.argmax(logits_except1, dim=1, keepdim=True)
    mask2 = one_hot(indices2_s, num_experts)
    gates1_s = (gates * mask1).sum(dim=1)
    gates2_s = (gates * mask2).sum(dim=1)

    if normalize_gate_prob_before_dropping:
        # Normalize gate probabilities
        denom_s = gates1_s + gates2_s
        # Avoid divide-by-zero
        denom_s = torch.clamp(denom_s, min=torch.finfo(denom_s.dtype).eps)
        gates1_s = gates1_s / denom_s
        gates2_s = gates2_s / denom_s

    if second_expert_policy == "random":
        sampled = (2 * gates2_s) > torch.rand_like(gates2_s)
        mask2 = mask2 * sampled.repeat(num_experts, 1).transpose(1, 0)

    # Compute locations in capacity buffer
    if input_mask is not None and input_mask.any():
        nonpadding = ~input_mask
        mask1 = mask1 * nonpadding.unsqueeze(-1).to(mask1.dtype)
        mask2 = mask2 * nonpadding.unsqueeze(-1).to(mask1.dtype)

    if batch_prioritized_routing:
        # if batch_prioritized_routing:
        importance_scores = -1 * gates.max(dim=1)[0]
        sorted_mask1 = mask1[importance_scores.argsort(dim=0)]
        sorted_cumsum1 = fused_cumsum_sub_one(sorted_mask1) * sorted_mask1
        importance_sorted_locations1 = sorted_cumsum1[
            importance_scores.argsort(dim=0).argsort(dim=0)
        ]

        sorted_mask2 = mask2[importance_scores.argsort(dim=0)]
        sorted_cumsum2 = fused_cumsum_sub_one(sorted_mask2) * sorted_mask2
        importance_sorted_locations2 = sorted_cumsum2[
            importance_scores.argsort(dim=0).argsort(dim=0)
        ]

        importance_sorted_locations2 += torch.sum(mask1, dim=0, keepdim=True)

        locations1, locations2 = (
            importance_sorted_locations1,
            importance_sorted_locations2,
        )
    else:
        locations1 = fused_cumsum_sub_one(mask1)
        locations2 = fused_cumsum_sub_one(mask2)
        # Update 2nd's location by accounting for locations of 1st
        locations2 += torch.sum(mask1, dim=0, keepdim=True)

    # Compute l_aux
    me = torch.mean(gates, dim=0)
    ce = torch.mean(mask1.to(gates.dtype), dim=0)
    l_aux = torch.mean(me * ce)
    l_aux = l_aux * num_experts * num_experts

    # for logging purposes
    metadata["overflow_expert1"] = (
        100
        * torch.sum(mask1 * torch.ge(locations1, capacity))
        / torch.sum(mask1)
    )
    metadata["overflow_expert2"] = (
        100
        * torch.sum(mask2 * torch.ge(locations2, capacity))
        / torch.sum(mask2)
    )

    # Remove locations outside capacity from mask
    mask1_, mask2_ = mask1, mask2
    mask1 = mask1 * torch.lt(locations1, capacity)
    mask2 = mask2 * torch.lt(locations2, capacity)

    # for logging (percent of tokens routed to each expert)
    expert1_hist = (
        100
        * torch.histc(
            (indices1_s.squeeze() + 1), bins=num_experts, min=1, max=num_experts
        )
        / num_tokens
    )
    metadata["unused_expert1_count"] = (expert1_hist == 0).sum()
    expert1_hist = (
        torch.sort(expert1_hist, dim=0, descending=True).values
        + torch.finfo(torch.float32).tiny
    )

    expert2_hist = (
        100
        * torch.histc(
            (indices2_s.squeeze() + 1), bins=num_experts, min=1, max=num_experts
        )
        / num_tokens
    )
    metadata["unused_expert2_count"] = (expert2_hist == 0).sum()
    expert2_hist = (
        torch.sort(expert2_hist, dim=0, descending=True).values
        + torch.finfo(torch.float32).tiny
    )

    sample_count = max(math.ceil(num_experts * SAMPLE_FRACTION), 1)
    metadata["expert1_balance_top"] = expert1_hist[:sample_count].sum()
    metadata["expert1_balance_bottom"] = expert1_hist[-sample_count:].sum()

    metadata["expert2_balance_top"] = expert2_hist[:sample_count].sum()
    metadata["expert2_balance_bottom"] = expert2_hist[-sample_count:].sum()

    if not normalize_gate_prob_before_dropping:
        # Normalize gate probabilities
        gates1_s = (gates * mask1).sum(dim=1)
        gates2_s = (gates * mask2).sum(dim=1)
        denom_s = gates1_s + gates2_s
        # Avoid divide-by-zero
        denom_s = torch.clamp(denom_s, min=torch.finfo(denom_s.dtype).eps)
        gates1_s /= denom_s
        gates2_s /= denom_s

    if has_tutel:
        locations1_s = torch.sum(locations1 * mask1_, dim=1)
        locations2_s = torch.sum(locations2 * mask2_, dim=1)
        return (
            l_aux,
            metadata,
            capacity,
            num_experts,
            [indices1_s, indices2_s],
            [locations1_s, locations2_s],
            [gates1_s, gates2_s],
        )

    # Store the capacity location for each token
    locations1_s = torch.sum(locations1 * mask1, dim=1)
    locations2_s = torch.sum(locations2 * mask2, dim=1)

    # Calculate combine_weights and dispatch_mask
    gates1 = gates1_s.unsqueeze(-1) * mask1.to(
        gates1_s.dtype
    )  # einsum("s,se->se")
    gates2 = gates2_s.unsqueeze(-1) * mask2.to(
        gates2_s.dtype
    )  # einsum("s,se->se")
    locations1_sc = one_hot(
        locations1_s, num_classes=capacity, unsqueeze_indices=True
    )
    locations2_sc = one_hot(
        locations2_s, num_classes=capacity, unsqueeze_indices=True
    )
    combine1_sec = torch.bmm(
        # einsum("se,sc->sec")
        gates1.unsqueeze(-1),
        locations1_sc.to(gates1.dtype).unsqueeze(1),
    )
    combine2_sec = torch.bmm(
        # einsum("se,sc->sec")
        gates2.unsqueeze(-1),
        locations2_sc.to(gates2.dtype).unsqueeze(1),
    )
    combine_weights = combine1_sec + combine2_sec
    dispatch_mask = combine_weights.bool()
    if use_fp32:
        return l_aux, combine_weights.to(orig_dtype), dispatch_mask, metadata
    else:
        return l_aux, combine_weights, dispatch_mask, metadata


class Top2Gate(torch.nn.Module):
    """Gate module which implements Top2Gating as described in Gshard_.
    ::

        gate = Top2Gate(model_dim, num_experts)
        l_aux, combine_weights, dispatch_mask = gate(input)

    .. Gshard_: https://arxiv.org/pdf/2006.16668.pdf

    Args:
        model_dim (int):
            size of model embedding dimension
        num_experts (ints):
            number of experts in model
    """

    wg: torch.nn.Linear

    def __init__(
        self,
        model_dim: int,
        num_experts: int,
        use_fp32=False,
        second_expert_policy="sampling",
        normalize_gate_prob_before_dropping=False,
        moe_eval_capacity_token_fraction=0.25,
        batch_prioritized_routing=False,
        use_xmoe=False,
    ) -> None:
        super().__init__()
        if not use_xmoe:
            self.wg = torch.nn.Linear(model_dim, num_experts, bias=False)
        else:
            self.wg_reduction = torch.nn.Linear(model_dim, 16, bias=False)
            wg = torch.empty(num_experts, 16)
            torch.nn.init.orthogonal_(wg, gain=0.32)
            self.register_parameter("wg", torch.nn.Parameter(wg))
        self.use_fp32 = use_fp32
        self.second_expert_policy = second_expert_policy
        self.normalize_gate_prob_before_dropping = (
            normalize_gate_prob_before_dropping
        )
        self.moe_eval_capacity_token_fraction = moe_eval_capacity_token_fraction
        self.batch_prioritized_routing = batch_prioritized_routing
        self.use_xmoe = use_xmoe

    def forward(self, input, mask=None):  # type: ignore
        if self.use_xmoe:
            input = self.wg_reduction(input)
            with torch.no_grad():
                wg_norm = self.wg.norm(p=2.0, dim=1, keepdim=True)
                self.wg.mul_(1.5 / wg_norm)
            logits = self._cosine(input, self.wg)
            logits = self._make_finite(logits)
        else:
            logits = self.wg(input)
        return top2gating(
            logits,
            mask,
            use_fp32=self.use_fp32,
            second_expert_policy=self.second_expert_policy,
            normalize_gate_prob_before_dropping=self.normalize_gate_prob_before_dropping,
            eval_mode=not self.training,
            moe_eval_capacity_token_fraction=self.moe_eval_capacity_token_fraction,
            batch_prioritized_routing=self.batch_prioritized_routing,
        )

    def _cosine(self, mat1, mat2, eps=1e-4):
        assert mat1.dim() == 2
        assert mat2.dim() == 2
        # mat1 = F.normalize(mat1, p=2.0, dim=1, eps=eps)
        mat2 = F.normalize(mat2.float(), p=2.0, dim=1, eps=eps)
        return mat1.float().matmul(mat2.transpose(0, 1)).type_as(mat1)

    def _make_finite(self, scores):
        ok = scores.isfinite()
        if not ok.all():
            # NaNs here can break the assignment algorithm
            scores[~ok] = scores[ok].min()
        return scores


# einsum dimensions: (g)roup, (s)equence, (e)xpert, (m)odel, (c)apacity


# Based on https://github.com/pytorch/pytorch/pull/40762
class _AllToAll(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, group: dist.ProcessGroup, input: Tensor) -> Tensor:  # type: ignore
        ctx.group = group
        input = input.contiguous()
        output = torch.empty_like(input)
        if torch.distributed.is_initialized():
            dist.all_to_all_single(output, input, group=group)
        else:
            assert group is None
            output = input
        return output

    @staticmethod
    def backward(ctx: Any, *grad_output: Tensor) -> Tuple[None, Tensor]:
        return (None, _AllToAll.apply(ctx.group, *grad_output))


class GShardMoELayer(Base):
    """
    Mixture of Experts (MOE) layer implementation.

    Args:
        gate (nn.Module): The gating network that determines the expert assignment.
        experts (Union[nn.ModuleList, nn.Module]): The expert networks.
        args (argparse.Namespace): The command-line arguments.

    Attributes:
        gate (nn.Module): The gating network that determines the expert assignment.
        experts (nn.ModuleList): The expert networks.
        expert_group (dist.ProcessGroup): The process group for experts.
        all2all_group (dist.ProcessGroup): The process group for all-to-all communication.
        world_size (int): The number of processes in the expert group.
        all2all_size (int): The number of processes in the all-to-all group.
        num_local_experts (int): The number of local experts.
        args (argparse.Namespace): The command-line arguments.
        in_generation (bool): Flag indicating if the layer is in generation mode.
        a2a_cuda_event_intervals (List[Tuple[torch.cuda.Event, torch.cuda.Event]]): List of CUDA event intervals for all-to-all communication.
        a2a_cpu_time_ms (float): Total CPU time spent on all-to-all communication.

    Methods:
        forward(*input: Tensor, input_padding_mask=None, **kwargs: Any) -> Tensor:
            Performs forward pass through the MOE layer.
        prepare_for_inference_():
            Prepares the MOE layer for inference mode.
        all_to_all_wrapper(input: Tensor) -> Tensor:
            Wrapper function for all-to-all communication.
        record_all_to_all_stats():
            Records statistics for all-to-all communication.
    """

    def __init__(self, gate, experts, args):
        if has_fairseq:
            super(Base, self).__init__()
        else:
            super().__init__()
        self.gate = gate
        if type(experts) == ModuleList:
            self.experts = cast(ModuleList, experts)
        else:
            self.experts = ModuleList([experts])
        _, self.expert_group = get_moe_group(args.moe_expert_count)
        self.all2all_group = get_all2all_group(args.moe_expert_count)
        self.world_size = dist.get_world_size(group=self.expert_group)
        self.all2all_size = dist.get_world_size(group=self.all2all_group)
        for p in experts.parameters():
            p.expert = True  # type: ignore
        self.num_local_experts = len(self.experts)
        self.args = args
        self.in_generation = False
        self.a2a_cuda_event_intervals = []
        self.a2a_cpu_time_ms = 0.0

    def forward(
        self, *input: Tensor, input_padding_mask=None, **kwargs: Any
    ) -> Tensor:
        assert len(input) == 1, "only single input Tensor supported"
        input = input[0]
        assert (
            len(input.shape) == 3
        ), "input Tensor must have dimensions: (s)equence, (t)oken, (m)odel"
        if input_padding_mask is not None:
            assert (
                len(input_padding_mask.shape) == 2
            ), "input Tensor must have dimensions: (s)equence, (t)oken"
            assert input_padding_mask.shape[0] == input.shape[0]
            assert input_padding_mask.shape[1] == input.shape[1]
        # assert input.shape[0] % len(self.experts) == 0, "num tokens must be order of number of local experts"

        # Implement Algorithm 2 from GShard paper.
        d_model = input.shape[2]
        # Pad to expected batch size
        input_shape = list(input.shape)
        expected_bsz = (
            getattr(self.args, "batch_size", 0)
            if self.training
            else getattr(self.args, "batch_size_valid", 0)
        )
        # This indicates that --batch-size or --max-sentences is not specified
        if expected_bsz is None:
            expected_bsz = 0
        # Note: Padding is not necessary at generation time at present
        # because all DDP workers process the same batch. Also, batch size at generation time
        # can be different from that present in the checkpoint state
        if (
            not self.in_generation
            and expected_bsz != 0
            and input_shape[0] != expected_bsz
        ):
            logger.warning(
                "padding batch with unexpected size"
                f" {input_shape[0]} (expected: {expected_bsz})"
            )
            assert (
                input_shape[0] < expected_bsz
            ), f"{input_shape[0]} < {expected_bsz}"
            padded_input = torch.zeros(
                (expected_bsz, input_shape[1], input_shape[2]),
                dtype=input.dtype,
                layout=input.layout,
                device=input.device,
            )
            padded_input[: input_shape[0], :, :] = input
            input = padded_input

            padded_input_padding_mask = torch.ones(
                (
                    expected_bsz,
                    input_shape[1],
                ),
                dtype=torch.bool,
                device=input.device,
            )
            if input_padding_mask is not None:
                padded_input_padding_mask[: input_shape[0], :] = (
                    input_padding_mask
                )
            else:
                padded_input_padding_mask[: input_shape[0], :] = False
            input_padding_mask = padded_input_padding_mask

        # Reshape into S tokens by dropping sequence dimension.
        reshaped_input = input.reshape(-1, d_model)
        reshaped_input_shape = reshaped_input.shape
        reshaped_input_padding_mask = (
            input_padding_mask.reshape(-1)
            if input_padding_mask is not None
            else None
        )

        # Doing padding here when --max-tokens is specified and not --batch-size or --max-sentences
        # Pro of --max-tokens: more flexible for MT variable sequence lengths
        # Con of --max-tokens: extra all-reduce needed to figure out optimal padding without running OOM
        if expected_bsz == 0:
            expected_dim = reshaped_input_shape[0] * torch.ones(
                (1,), dtype=torch.long, device=input.device
            )
            dist.all_reduce(
                expected_dim, group=dist.group.WORLD, op=dist.ReduceOp.MAX
            )
            expected_dim = int(expected_dim.item())
            padded_input = torch.zeros(
                (expected_dim, reshaped_input_shape[1]),
                dtype=input.dtype,
                layout=input.layout,
                device=input.device,
            )
            padded_input[: reshaped_input_shape[0], :] = reshaped_input
            reshaped_input = padded_input

            padded_input_padding_mask = torch.ones(
                (expected_dim,), dtype=torch.bool, device=padded_input.device
            )
            if reshaped_input_padding_mask is not None:
                padded_input_padding_mask[: reshaped_input_shape[0]] = (
                    reshaped_input_padding_mask
                )
            else:
                padded_input_padding_mask[: reshaped_input_shape[0]] = False
            reshaped_input_padding_mask = padded_input_padding_mask

        if has_tutel:
            (
                l_aux,
                self.metadata,
                C,
                E,
                indices_,
                locations_,
                gates_,
            ) = self.gate(reshaped_input, reshaped_input_padding_mask)
            S, M = reshaped_input.size(0), reshaped_input.size(1)

            if not hasattr(self, "_tutel_dispatcher"):
                self._tutel_dispatcher = tutel_moe.fast_dispatcher(
                    E, C, M, dispatch_dtype=reshaped_input.dtype
                )
            self._tutel_dispatcher.update(
                indices_, locations_, gates_, capacity=C
            )
            dispatched_input = self._tutel_dispatcher.encode(reshaped_input)
        else:
            l_aux, combine_weights, dispatch_mask, self.metadata = self.gate(
                reshaped_input, reshaped_input_padding_mask
            )

            dispatch_mask = dispatch_mask.to(input.dtype).permute(
                1, 2, 0
            )  # S,E,C -> E,C,S
            E, C, S = dispatch_mask.size()
            M = reshaped_input.size(1)
            assert reshaped_input.size() == (S, M)
            # einsum("sec,sm->ecm")
            dispatched_input = torch.mm(
                dispatch_mask.view(E * C, S), reshaped_input
            )  # -> (E*C),M

        if self.all2all_size > 1:
            dispatched_input = self.all_to_all_wrapper(dispatched_input)

        # Re-shape after all-to-all: ecm -> gecm
        dispatched_input = dispatched_input.reshape(
            self.all2all_size, self.num_local_experts, -1, d_model
        )
        chunks = dispatched_input.chunk(self.num_local_experts, dim=1)
        expert_outputs = []
        for chunk, expert in zip(chunks, self.experts):
            expert_outputs += [expert(chunk)]
        expert_output = torch.cat(expert_outputs, dim=1)

        if self.all2all_size > 1:
            expert_output = self.all_to_all_wrapper(expert_output)

        # Re-shape back: gecm -> ecm
        expert_output = expert_output.reshape(
            self.all2all_size * self.num_local_experts, -1, d_model
        )

        if has_tutel:
            combined_output = self._tutel_dispatcher.decode(
                expert_output.view(E * C, M)
            )
        else:
            # einsum("sec,ecm->sm")
            combined_output = combine_weights.view(S, E * C).mm(
                expert_output.view(E * C, M)
            )

        # Remove padding here when --max-tokens is specified and not --batch-size or --max-sentences
        combined_output = combined_output[: reshaped_input_shape[0], :]
        combined_output = combined_output.reshape(input.shape)
        combined_output = combined_output[: input_shape[0], :, :]

        self.record_all_to_all_stats()

        return combined_output, l_aux

    def prepare_for_inference_(self):
        self.in_generation = True

    def all_to_all_wrapper(self, input: Tensor):
        dummy_a2a = getattr(self.args, "dummy_a2a", False)
        if dummy_a2a:
            input = input.contiguous()
            output = input.detach().clone()
            return input
        # always record times, since it is not a lot of overhead
        # if we do not log it we simply clear it off in record_all_to_all_stats
        cuda_start = torch.cuda.Event(enable_timing=True)
        cuda_end = torch.cuda.Event(enable_timing=True)
        cpu_start = time.time() * 1000
        cuda_start.record()
        output = _AllToAll.apply(self.all2all_group, input)
        cuda_end.record()
        cpu_end = time.time() * 1000
        self.a2a_cpu_time_ms += cpu_end - cpu_start
        self.a2a_cuda_event_intervals.append((cuda_start, cuda_end))
        return output

    def record_all_to_all_stats(self):
        # controlled via an argument as we want to minimize any impact from torch.cuda.synchronize()
        record_a2a_perf_stats = getattr(
            self.args, "record_a2a_perf_stats", False
        )
        if record_a2a_perf_stats:
            torch.cuda.synchronize()
            self.metadata["all_to_all_cpu_time_ms"] = self.a2a_cpu_time_ms
            a2a_cuda_time_ms = 0.0
            for ev_start, ev_end in self.a2a_cuda_event_intervals:
                a2a_cuda_time_ms += ev_start.elapsed_time(ev_end)
            self.metadata["all_to_all_cuda_time_ms"] = a2a_cuda_time_ms
        # reset stats
        self.a2a_cpu_time_ms = 0.0
        self.a2a_cuda_event_intervals = []



from torch import nn, Tensor


class VerboseExecution(nn.Module):
    """
    A wrapper class that adds verbosity to the execution of a given model.

    Args:
        model (nn.Module): The model to be executed.
    """

    def __init__(self, model: nn.Module):
        super().__init__()
        self.model = model

        for name, layer in self.model.named_children():
            for name, layer in self.model.named_children():
                layer.__name__ = name
                layer.register_forward_hook(
                    lambda layer, _, output: print(
                        f"{layer.__name__} output: {output.shape}"
                    )
                )

    def forward(self, x: Tensor) -> Tensor:
        return self.model(x)



import torch
from torch import nn


class Exo(nn.Module):
    """

    Exo activation function
    -----------------------

    Exo is a new activation function that is a combination of linear and non-linear parts.

    Formula
    -------
    .. math::
        f(x) = \\sigma(x) \\cdot x + (1 - \\sigma(x)) \\cdot tanh(x)

    Parameters
    ----------
    alpha : float
        Alpha value for the activation function. Default: 1.0

    Examples
    --------
    >>> m = Exo()
    >>> input = torch.randn(2)
    >>> output = m(input)


    # Paper

    **"Exo": A Conceptual Framework**
    For the sake of this exercise, let's envision Exo as a cutting-edge activation function
    inspired by the idea of "extraterrestrial" or "outside the norm" data processing. The main premise is that it's designed to handle the vast heterogeneity in multi-modal data by dynamically adjusting its transformation based on the input distribution.

    ---

    **Technical Report on the Exo Activation Function**

    **Abstract**

    In the evolving landscape of deep learning and multi-modal data processing,
    activation functions play a pivotal role. This report introduces the "Exo" activation
     function, a novel approach designed to cater to the diverse challenges posed by multi-modal data.
    Rooted in a dynamic mechanism, Exo adjusts its transformation based on the input distribution, offering flexibility and efficiency in handling heterogeneous data.

    **1. Introduction**

    Activation functions serve as the heart of neural networks, determining the output of a node given an input or set of inputs. As deep learning models grow in complexity, especially in the realm of multi-modal data processing, there's a pressing need for activation functions that are both versatile and computationally efficient. Enter Exo—a dynamic, adaptive function envisioned to rise to this challenge.

    **2. Design Philosophy**

    At its core, Exo embodies the idea of adaptability. Drawing inspiration from the vast, unpredictable expanse of outer space, Exo is designed to dynamically adjust to the data it processes. This inherent flexibility makes it a prime candidate for multi-modal tasks, where data can be as varied as the stars in the sky.

    **3. Mechanism of Operation**

    Exo operates on a simple yet powerful principle: adaptive transformation. It leverages a gating mechanism that weighs the influence of linear versus non-linear transformations based on the magnitude and distribution of input data.

    The pseudocode for Exo is as follows:

    ```
    function Exo(x, alpha):
        gate = sigmoid(alpha * x)
        linear_part = x
        non_linear_part = tanh(x)
        return gate * linear_part + (1 - gate) * non_linear_part
    ```

    **4. Why Exo Works the Way It Does**

    The strength of Exo lies in its adaptive nature. The gating mechanism—dictated by
    the sigmoid function—acts as a switch. For high-magnitude inputs,
    Exo trends towards a linear behavior. Conversely, for lower-magnitude inputs,
    it adopts a non-linear transformation via the tanh function.

    This adaptability allows Exo to efficiently handle data heterogeneity,
    a prominent challenge in multi-modal tasks.

    **5. Ideal Use Cases**

    Given its versatile nature, Exo shows promise in the following domains:

    - **Multi-Modal Data Processing**: Exo's adaptability makes it a strong contender
    for models handling diverse data types, be it text, image, or audio.

    - **Transfer Learning**: The dynamic range of Exo can be beneficial when transferring
    knowledge from one domain to another.

    - **Real-time Data Streams**: For applications where data distributions might change
    over time, Exo's adaptive nature can offer robust performance.

    **6. Experimental Evaluation**

    Future research will rigorously evaluate Exo against traditional activation functions
    across varied datasets and tasks.

    ---

    **Methods Section for a Research Paper**

    **Methods**

    **Activation Function Design**

    The Exo activation function is defined as:

    \[ Exo(x) = \sigma(\alpha x) \times x + (1 - \sigma(\alpha x)) \times \tanh(x) \]

    where \(\sigma\) represents the sigmoid function, and \(\alpha\) is a hyperparameter
    dictating the sensitivity of the gating mechanism.

    **Model Configuration**

    All models were built using the same architecture, with the only difference
    being the activation function. This ensured that any performance disparities
    were solely attributed to the activation function and not other model parameters.

    **Datasets and Pre-processing**

    Three diverse datasets representing image, text, and audio modalities were employed.
    All datasets underwent standard normalization procedures.

    **Training Regimen**

    Models were trained using the Adam optimizer with a learning rate of 0.001 for
    50 epochs. Performance metrics, including accuracy and loss, were recorded.


    """

    def __init__(self, alpha=1.0):
        """INIT function."""
        super(Exo, self).__init__()

    def forward(self, x):
        """Forward function."""
        gate = torch.sigmoid(x)
        linear_part = x
        non_linear_part = torch.tanh(x)
        return gate * linear_part + (1 - gate) * non_linear_part



import torch
from torch import nn
from einops.layers.torch import Rearrange
import torch.nn.functional as F


class SpatialTransformer(nn.Module):
    """
    Spacial Transformer Network

    https://pytorch.org/tutorials/intermediate/spatial_transformer_tutorial.html

    Usage:
    >>> stn = SpatialTransformer()
    >>> stn.stn(x)

    """

    def __init__(self):
        super(SpatialTransformer, self).__init__()

        # spatial transformer localization-network
        linear = nn.Linear(32, 3 * 2)

        # initialize the weights/bias with identity transformation
        linear.weight.data.zero_()

        linear.bias.data.copy_(
            torch.tensor([1, 0, 0, 0, 1, 0], dtype=torch.float)
        )

        self.compute_theta = nn.Sequential(
            nn.Conv2d(1, 8, kernel_size=7),
            nn.MaxPool2d(2, stride=2),
            nn.ReLU(True),
            nn.Conv2d(8, 10, kernel_size=5),
            nn.MaxPool2d(2, stride=2),
            nn.ReLU(True),
            Rearrange("b c h w -> b (c h w)", h=3, w=3),
            nn.Linear(10 * 3 * 3, 32),
            nn.ReLU(True),
            linear,
            Rearrange("b (row col) -> b row col", row=2, col=3),
        )

    def stn(self, x):
        """
        stn module
        """
        grid = F.affine_grid(self.compute_theta(x), x.size())
        return F.grid_sample(x, grid)



from __future__ import annotations

import math

import torch
import torch.nn.functional as F
from einops import einsum, rearrange, repeat
from torch import Tensor, nn

from zeta.nn.modules.rms_norm import RMSNorm
from zeta.utils import exists


class MambaBlock(nn.Module):
    """
    Initialize a single Mamba block.

    Args:
        dim (int): The input dimension.
        dim_inner (Optional[int]): The inner dimension. If not provided, it is set to dim * expand.
        depth (int): The depth of the Mamba block.
        d_state (int): The state dimension. Default is 16.
        expand (int): The expansion factor. Default is 2.
        dt_rank (Union[int, str]): The rank of the temporal difference (Δ) tensor. Default is "auto".
        d_conv (int): The dimension of the convolutional kernel. Default is 4.
        conv_bias (bool): Whether to include bias in the convolutional layer. Default is True.
        bias (bool): Whether to include bias in the linear layers. Default is False.

    Examples:
        >>> import torch
        >>> from zeta.nn.modules.simple_mamba import MambaBlock
        >>> block = MambaBlock(dim=64, depth=1)
        >>> x = torch.randn(1, 10, 64)
        >>> y = block(x)
        >>> y.shape
        torch.Size([1, 10, 64])
    """

    def __init__(
        self,
        dim: int = None,
        depth: int = 5,
        d_state: int = 16,
        expand: int = 2,
        d_conv: int = 4,
        conv_bias: bool = True,
        bias: bool = False,
    ):
        """A single Mamba block, as described in Figure 3 in Section 3.4 in the Mamba paper [1]."""
        super().__init__()
        self.dim = dim
        self.depth = depth
        self.d_state = d_state
        self.expand = expand
        self.d_conv = d_conv
        self.conv_bias = conv_bias
        self.bias = bias

        # If dt_rank is not provided, set it to ceil(dim / d_state)
        dt_rank = math.ceil(self.dim / 16)
        self.dt_rank = dt_rank

        # If dim_inner is not provided, set it to dim * expand
        dim_inner = dim * expand
        self.dim_inner = dim_inner

        # If dim_inner is not provided, set it to dim * expand
        self.in_proj = nn.Linear(dim, dim_inner * 2, bias=bias)

        self.conv1d = nn.Conv1d(
            in_channels=dim_inner,
            out_channels=dim_inner,
            bias=conv_bias,
            kernel_size=d_conv,
            groups=dim_inner,
            padding=d_conv - 1,
        )

        # x_proj takes in `x` and outputs the input-specific Δ, B, C
        self.x_proj = nn.Linear(
            dim_inner, dt_rank + self.d_state * 2, bias=False
        )

        # dt_proj projects Δ from dt_rank to d_in
        self.dt_proj = nn.Linear(dt_rank, dim_inner, bias=True)

        A = repeat(torch.arange(1, self.d_state + 1), "n -> d n", d=dim_inner)
        self.A_log = nn.Parameter(torch.log(A))
        self.D = nn.Parameter(torch.ones(dim_inner))
        self.out_proj = nn.Linear(dim_inner, dim, bias=bias)

    def forward(self, x: Tensor):
        """Mamba block forward. This looks the same as Figure 3 in Section 3.4 in the Mamba paper [1].

        Args:
            x: shape (b, l, d)    (See Glossary at top for definitions of b, l, d_in, n...)

        Returns:
            output: shape (b, l, d)


        Official Implementation:
            class Mamba, https://github.com/state-spaces/mamba/blob/main/mamba_ssm/modules/mamba_simple.py#L119
            mamba_inner_ref(), https://github.com/state-spaces/mamba/blob/main/mamba_ssm/ops/selective_scan_interface.py#L311

        """
        (b, l, d) = x.shape

        x_and_res = self.in_proj(x)  # shape (b, l, 2 * d_in)
        x_and_res = rearrange(x_and_res, "b l x -> b x l")
        (x, res) = x_and_res.split(
            split_size=[self.dim_inner, self.dim_inner], dim=1
        )

        x = self.conv1d(x)[:, :, :l]
        x = F.silu(x)

        y = self.ssm(x)

        y = y * F.silu(res)

        output = self.out_proj(rearrange(y, "b dim l -> b l dim"))

        return output

    def ssm(self, x: Tensor):
        """Runs the SSM. See:
            - Algorithm 2 in Section 3.2 in the Mamba paper [1]
            - run_SSM(A, B, C, u) in The Annotated S4 [2]

        Args:
            x: shape (b, d_in, l)    (See Glossary at top for definitions of b, l, d_in, n...)

        Returns:
            output: shape (b, d_in, l)

        Official Implementation:
            mamba_inner_ref(), https://github.com/state-spaces/mamba/blob/main/mamba_ssm/ops/selective_scan_interface.py#L311

        """
        (d_in, n) = self.A_log.shape

        # Compute ∆ A B C D, the state space parameters.
        #     A, D are input independent
        #     ∆, B, C are input-dependent (this is a key difference between Mamba and the linear time invariant S4)

        A = -torch.exp(self.A_log.float())  # shape (d_in, n)
        D = self.D.float()

        x_dbl = rearrange(x, "b d l -> b l d")
        x_dbl = self.x_proj(x_dbl)  # (b, l, dt_rank + 2*n)

        (delta, B, C) = x_dbl.split(
            split_size=[self.dt_rank, n, n], dim=-1
        )  # delta: (b, l, dt_rank). B, C: (b, l, n)
        delta = F.softplus(self.dt_proj(delta))  # (b, l, d_in)

        y = self.selective_scan(
            x, delta, A, B, C, D
        )  # This is similar to run_SSM(A, B, C, u) in The Annotated S4 [2]

        return y

    def selective_scan(self, u, delta, A, B, C, D):
        """Does selective scan algorithm. See:
            - Section 2 State Space Models in the Mamba paper [1]
            - Algorithm 2 in Section 3.2 in the Mamba paper [1]
            - run_SSM(A, B, C, u) in The Annotated S4 [2]

        This is the classic discrete state space formula:
            x(t + 1) = Ax(t) + Bu(t)
            y(t)     = Cx(t) + Du(t)
        except B and C (and the step size delta, which is used for discretization) are dependent on the input x(t).

        Args:
            u: shape (b, d_in, l)    (See Glossary at top for definitions of b, l, d_in, n...)
            delta: shape (b, l, d_in)
            A: shape (d_in, n)
            B: shape (b, l, n)
            C: shape (b, l, n)
            D: shape (d_in,)

        Returns:
            output: shape (b, d_in, l)

        Official Implementation:
            selective_scan_ref(), https://github.com/state-spaces/mamba/blob/main/mamba_ssm/ops/selective_scan_interface.py#L86
            Note: I refactored some parts out of `selective_scan_ref` out, so the functionality doesn't match exactly.

        """
        (b, d_in, l) = u.shape
        n = A.shape[1]

        # Discretize continuous parameters (Δ, A, B)  (see Section 2 Equation 4 in the Mamba paper [1])
        # Note that B is parameterized directly
        deltaA = torch.exp(einsum(delta, A, "b l d_in, d_in n -> b d_in l n"))
        deltaB_u = einsum(
            delta, B, u, "b l d_in, b l n, b d_in l -> b d_in l n"
        )

        # Perform selective scan (see scan_SSM() in The Annotated S4 [2])
        x = torch.zeros((b, d_in, n))
        ys = []
        for i in range(l):
            x = deltaA[:, :, i] * x + deltaB_u[:, :, i]
            y = einsum(x, C[:, i, :], "b d_in n , b n -> b d_in")
            ys.append(y)
        y = torch.stack(ys, dim=2)  # (b d_in l)

        if D is not None:
            y = y + u * rearrange(D, "d_in -> d_in 1")

        return y


class Mamba(nn.Module):
    """Mamba model.

    Args:
        vocab_size (int): The size of the vocabulary.
        dim (int): The input dimension.
        depth (int): The depth of the Mamba block.
        d_state (int): The state dimension. Default is 16.
        expand (int): The expansion factor. Default is 2.
        dt_rank (Union[int, str]): The rank of the temporal difference (Δ) tensor. Default is "auto".
        d_conv (int): The dimension of the convolutional kernel. Default is 4.

    Examples:
    x = torch.randint(0, 16, (1, 64))
    model = Mamba(16, 64, 5, 16)
    out = model(x)
    print(out)
    """

    def __init__(
        self,
        vocab_size: int = None,
        dim: int = None,
        depth: int = 5,
        d_state: int = 16,
        img_dim: int = 64,
        *args,
        **kwargs,
    ):
        """Full Mamba model."""
        super().__init__()

        self.embedding = nn.Embedding(vocab_size, dim)
        self.norm_f = RMSNorm(dim)
        self.lm_head = nn.Linear(dim, vocab_size, bias=False)
        self.lm_head.weight = self.embedding.weight
        self.mamba_layers = nn.ModuleList(
            [
                MambaBlock(
                    dim=dim, depth=depth, d_state=d_state, *args, **kwargs
                )
                for _ in range(depth)
            ]
        )

        # Projection for img
        self.img_proj = nn.Linear(img_dim, dim)

    def forward(
        self,
        x: Tensor,
        context: Tensor = None,
    ):
        """
        Args:
            x (long tensor): shape (b, l)    (See Glossary at top for definitions of b, l, d_in, n...)

        Returns:
            logits: shape (b, l, vocab_size)

        Official Implementation:
            class MambaLMHeadModel, https://github.com/state-spaces/mamba/blob/main/mamba_ssm/models/mixer_seq_simple.py#L173

        """
        x = self.embedding(x)

        if exists(context):
            # Project the image
            projected_img = self.img_proj(context)

            # Concatenate the image and text
            x = torch.cat([x, projected_img], dim=1)

        for layer in self.mamba_layers:
            x = layer(self.norm_f(x)) + x

        x = self.norm_f(x)
        logits = self.lm_head(x)

        return logits



from torch import nn, Tensor
from typing import Optional
from einops import rearrange
from zeta.nn.modules.ssm import SSM


class VSSBlock(nn.Module):
    """
    VSSBlock is a module that implements a Variational State Space (VSS) block.

    PAPER: https://arxiv.org/pdf/2401.10166.pdf

    Args:
        dim (int): The input dimension.
        d_state (int): The dimension of the state.
        dim_head (int): The dimension of each head in the multi-head attention mechanism.
        heads (int): The number of attention heads.
        dt_rank (int): The rank of the dynamic tensor.
        dim_inner (Optional[int]): The inner dimension of the feed-forward network. Defaults to None.

    Attributes:
        dim (int): The input dimension.
        d_state (int): The dimension of the state.
        dim_head (int): The dimension of each head in the multi-head attention mechanism.
        heads (int): The number of attention heads.
        dt_rank (int): The rank of the dynamic tensor.
        dim_inner (int): The inner dimension of the feed-forward network.
        scale (float): The scaling factor for the attention weights.
        norm (nn.LayerNorm): The layer normalization module.
        depthwise_conv (nn.Conv1d): The depthwise convolution layer.
        proj (nn.Linear): The linear projection layer.
        ssm (SSM): The Variational State Space Model (SSM) module.

    """

    def __init__(
        self,
        dim: int,
        d_state: int,
        dim_head: int,
        heads: int,
        dt_rank: int,
        dim_inner: Optional[int] = None,
    ):
        super().__init__()
        self.dim = dim
        self.d_state = d_state
        self.dim_head = dim_head
        self.heads = heads
        self.dt_rank = dt_rank
        self.dim_inner = dim_inner if dim_inner is not None else dim * 4

        self.scale = dim_head**-0.5

        self.norm = nn.LayerNorm(dim)
        self.depthwise_conv = nn.Conv1d(
            dim,
            dim,
            kernel_size=3,
            padding=1,
        )
        self.proj = nn.Linear(dim, dim)
        self.ssm = SSM(
            in_features=dim,
            dt_rank=dt_rank,
            dim_inner=dim_inner,
            d_state=d_state,
        )

    def forward(self, x: Tensor):
        """
        Forward pass of the VSSBlock module.

        Args:
            x (Tensor): Input tensor.

        Returns:
            Tensor: Output tensor after passing through the VSSBlock module.
        """
        skip = x

        x = self.norm(x)

        # Linear projection
        x = self.proj(x)

        linear_skip = x
        linear_skip = self.proj(linear_skip)

        # Depthwise convolution
        x = rearrange(x, "b n (h d) -> b (n h) d", h=self.heads)
        x = self.depthwise_conv(x)
        x = rearrange(x, "b (n h) d -> b n (h d)", h=self.heads)

        # SSM
        x = self.ssm(x)

        # Layernorm
        x = self.norm(x)

        # Matmul with layernorm and skip connection
        x = x @ linear_skip

        # linear
        x = self.proj(x)

        # Addition
        x + skip



import torch
from torch import nn
from torch import abs, softmax, sqrt, tensor, topk


class SparQAttention(nn.Module):
    """
    Sparse and Quantized Attention (SparQAttention) is a novel attention mechanism
    that approximates the attention scores using the r largest components of the query matrix
    and then gathers the top k positions based on the approximate attention scores.


    Methods:
        forward(Q, K, V, V_mean, M, r, k): Computes the Sparse and Quantized attention.

    Examples:
    >>> import torch
    >>> from zeta.nn.modules import SparQAttention
    >>> attention = SparQAttention()
    >>> batch_size, heads, seq_length, dim = 2, 4, 10, 64
    >>> Q = torch.randn(batch_size, heads, seq_length, dim)
    >>> K = torch.randn(batch_size, heads, seq_length, dim)
    >>> V = torch.randn(batch_size, heads, seq_length, dim)
    >>> V_mean = torch.randn(batch_size, heads, 1, dim)
    >>> M = torch.randn(batch_size, heads, seq_length, seq_length)
    >>> r = 5  # Number of largest components for approximation
    >>> k = 5  # Number of top positions for attention
    >>> output = attention.forward(Q, K, V, V_mean, M, r, k)
    >>> print(output)




    """

    def __init__(self, dim: int = None, heads: int = None, *args, **kwargs):
        """Initialize the SparQAttention class."""
        super().__init__(*args, **kwargs)
        self.dim = dim
        self.heads = heads

    def forward(
        self,
        Q: torch.Tensor,
        K: torch.Tensor,
        V: torch.Tensor,
        V_mean: torch.Tensor,
        M: torch.Tensor,
        r: int,
        k: int,
        *args,
        **kwargs,
    ):
        """
        Computes the Sparse and Quantized attention.

        Args:
            Q (Tensor): Query matrix.
            K (Tensor): Key matrix.
            V (Tensor): Value matrix.
            V_mean (Tensor): Mean of values.
            M (Tensor): Mask.
            r (int): Number of largest components for approximation.
            k (int): Number of top positions for attention.

        Returns:
            Tensor: The result of applying sparse quantized attention.
        """
        try:
            # # Make sure that the input tensors match the specified dimensions
            # assert Q.size(1) == self.heads and Q.size(-1) == self.dim, \
            #     "Query tensor dimensions do not match the specified number of heads and head dimension"
            # assert K.size(1) == self.heads and K.size(-1) == self.dim, \
            #     "Key tensor dimensions do not match the specified number of heads and head dimension"
            # assert V.size(1) == self.heads and V.size(-1) == self.dim, \
            #     "Value tensor dimensions do not match the specified number of heads and head dimension"

            # Gather function
            def gather(t, dim, i):
                dim += (dim < 0) * t.dim()
                return t.gather(
                    dim,
                    i.expand(*t.shape[:dim], i.shape[dim], *t.shape[dim + 1 :]),
                )

            # Attention function
            def attn(q, k, v, m):
                s = q @ k.transpose(-1, -2) / sqrt(tensor(q.shape[-1])) + m
                return softmax(s, dim=-1) @ v

            # 1. Approximate attention scores using r largest components of Q
            i1 = topk(abs(Q), r, -1).indices
            Q_hat, K_hat = gather(Q, -1, i1), gather(K, -1, i1)
            scale = sqrt(
                Q.shape[-1]
                * abs(Q_hat).sum(dim=-1, keepdim=True)
                / abs(Q).sum(dim=-1, keepdim=True)
            )
            s_hat = softmax(Q_hat @ K_hat.transpose(-1, -2) / scale + M, dim=-1)

            # 2. Gather top k positions based on approximate attention scores & run attention
            i2 = topk(s_hat, k, -1).indices
            iKV = i2[..., 0, :, None]
            K, V, M = gather(K, -2, iKV), gather(V, -2, iKV), gather(M, -1, i2)
            y_ = attn(Q, K, V, M)

            # 3. Estimate the total score of the top k, and interpolate with V_mean
            alpha = gather(s_hat, -1, i2).sum(-1, keepdim=True)
            return alpha * y_ + (1 - alpha) * V_mean
        except Exception as e:
            raise ValueError(f"Error in SPARQ attention computation: {e}")


# Example usage
num_heads = 4
head_dim = 64
attention = SparQAttention(num_heads, head_dim)

# Generate random tensors with the specified dimensions
batch_size, seq_length = 2, 10
Q = torch.randn(batch_size, num_heads, seq_length, head_dim)
K = torch.randn(batch_size, num_heads, seq_length, head_dim)
V = torch.randn(batch_size, num_heads, seq_length, head_dim)
V_mean = torch.randn(batch_size, num_heads, 1, head_dim)
M = torch.randn(batch_size, num_heads, seq_length, seq_length)

# Compute the Sparse and Quantized attention
r = 5  # Number of largest components for approximation
k = 5  # Number of top positions for attention
output = attention.forward(Q, K, V, V_mean, M, r, k)

# Output tensor
print(output)



import torch
from torch import nn, Tensor
from einops import rearrange
from math import sqrt


class DepthWiseConv2d(nn.Module):
    def __init__(
        self, dim_in, dim_out, kernel_size, padding, stride, bias=True
    ):
        super().__init__()
        self.net = nn.Sequential(
            nn.Conv2d(
                dim_in,
                dim_out,
                kernel_size=kernel_size,
                padding=padding,
                groups=dim_in,
                stride=stride,
                bias=bias,
            ),
            nn.Conv2d(dim_out, dim_out, kernel_size=1, bias=bias),
        )

    def forward(self, x):
        return self.net(x)


# pooling layer


class Pool(nn.Module):
    def __init__(self, dim: int):
        """
        Pool module that performs pooling operation on input tensors.

        Args:
            dim (int): The input tensor dimension.

        """
        super().__init__()
        self.downsample = DepthWiseConv2d(
            dim, dim * 2, kernel_size=3, stride=2, padding=1
        )
        self.cls_ff = nn.Linear(dim, dim * 2)

    def forward(self, x: Tensor):
        """
        Forward pass of the Pool module.

        Args:
            x (Tensor): The input tensor.

        Returns:
            Tensor: The output tensor after pooling operation.

        """
        cls_token, tokens = x[:, :1], x[:, 1:]
        cls_token = self.cls_ff(cls_token)
        tokens = rearrange(
            tokens, "b (h w) c -> b c h w", h=int(sqrt(tokens.shape[1]))
        )
        tokens = self.downsample(tokens)
        tokens = rearrange(tokens, "b c h w -> b (h w) c")
        return torch.cat((cls_token, tokens), dim=1)



import torch
from accelerate import Accelerator

from transformers import (
    get_cosine_schedule_with_warmup,
    get_linear_schedule_with_warmup,
)


def get_lr_scheduler_with_warmup(
    optimizer: torch.optim.Optimizer,
    scheduler_type: str,
    num_warmup_steps: int,
    max_train_steps: int,
    grad_accumulate_every: int = 1,
    accelerator: Accelerator = None,
):
    """
    Get a learning rate scheduler with warmup.

    Args:
        optimizer (Optimizer): The optimizer for which to create the learning rate scheduler.
        scheduler_type (str): The type of learning rate scheduler to create, either "linear" or "cosine".
        num_warmup_steps (int): The number of warmup steps for the learning rate scheduler.
        max_train_steps (int): The maximum number of training steps.
        grad_accumulate_every (int, optional): The gradient accumulation factor. Defaults to 1.
        accelerator (Accelerator, optional): The Accelerate library accelerator. Defaults to None.

    Returns:
        The learning rate scheduler with warmup.

    Raises:
        ValueError: If scheduler_type is not "linear" or "cosine".
    """
    NUM_WARMUP_STEPS = num_warmup_steps
    GRADIENT_ACCUMULATE_EVERY = grad_accumulate_every
    if accelerator is not None:
        accelerator.print(f"Using {scheduler_type} lr scheduler")
    if scheduler_type == "linear":
        return get_linear_schedule_with_warmup(
            optimizer=optimizer,
            num_warmup_steps=NUM_WARMUP_STEPS * GRADIENT_ACCUMULATE_EVERY,
            num_training_steps=max_train_steps * GRADIENT_ACCUMULATE_EVERY,
        )
    elif scheduler_type == "cosine":
        return get_cosine_schedule_with_warmup(
            optimizer=optimizer,
            num_warmup_steps=NUM_WARMUP_STEPS * GRADIENT_ACCUMULATE_EVERY,
            num_training_steps=max_train_steps * GRADIENT_ACCUMULATE_EVERY,
        )
    else:
        raise ValueError(
            "Invalid scheduler_type. Expected 'linear' or 'cosine', got: {}"
            .format(scheduler_type)
        )



from einops import rearrange
import torch
from torch import nn
from zeta.nn.modules.ssm import SSM


class VisionMambaBlock(nn.Module):
    """
    VisionMambaBlock is a module that implements the Mamba block from the paper
    Vision Mamba: Efficient Visual Representation Learning with Bidirectional
    State Space Model

    Args:
        dim (int): The input dimension of the input tensor.
        heads (int): The number of heads in the multi-head attention mechanism.
        dt_rank (int): The rank of the state space model.
        dim_inner (int): The dimension of the inner layer of the multi-head attention.
        d_state (int): The dimension of the state space model.


    Example:
    >>> block = VisionMambaBlock(dim=256, heads=8, dt_rank=32, dim_inner=512, d_state=256)
    >>> x = torch.randn(1, 32, 256)
    >>> out = block(x)
    >>> out.shape
    torch.Size([1, 32, 256])
    """

    def __init__(
        self, dim: int, heads: int, dt_rank: int, dim_inner: int, d_state: int
    ):
        super().__init__()
        self.dim = dim
        self.heads = heads
        self.dt_rank = dt_rank
        self.dim_inner = dim_inner
        self.d_state = d_state

        self.forward_conv1d = nn.Conv1d(
            in_channels=dim, out_channels=dim, kernel_size=1
        )
        self.backward_conv1d = nn.Conv1d(
            in_channels=dim, out_channels=dim, kernel_size=1
        )
        self.norm = nn.LayerNorm(dim)
        self.activation = nn.SiLU()
        self.ssm = SSM(dim, dt_rank, dim_inner, d_state)

    def forward(self, x: torch.Tensor):
        """Forward pass of the VisionMambaBlock module.

        Args:
            x (torch.Tensor): _description_

        Returns:
            _type_: _description_
        """
        # x is of shape [batch_size, seq_len, dim]
        # Use einops to rearrange for Conv1d
        skip = x
        x = self.norm(x)

        z1 = x
        x1 = x

        # forward con1d
        x1_rearranged = rearrange(x1, "b s d -> b d s")
        forward_conv_output = self.forward_conv1d(x1_rearranged)
        forward_conv_output = rearrange(forward_conv_output, "b d s -> b s d")
        x1_ssm = self.ssm(forward_conv_output)

        # backward conv x2
        x2_rearranged = rearrange(x1, "b s d -> b d s")
        x2 = self.backward_conv1d(x2_rearranged)
        x2 = rearrange(x2, "b d s -> b s d")

        # Backward ssm
        x2 = self.ssm(x2)

        # Activation
        z = self.activation(z1)

        # matmul with z + backward ssm
        x2 = x2 @ z

        # Matmul with z and x1
        x1 = x1_ssm @ z

        # Add both matmuls
        x = x1 + x2

        # Add skip connection
        return x + skip



import math
from typing import Optional

from torch import Tensor, nn


class Conv2DFeedforward(nn.Module):
    """
    A Convolutional feed-forward network, as proposed in VAN_ (Vision Attention Network, Guo et al.)

    .. _VAN: https://arxiv.org/pdf/2202.09741.pdf


    Example::

        >>> import torch
        >>> from zeta.nn import Conv2DFeedforward
        >>> m = Conv2DFeedforward(256, 1, 256)
        >>> x = torch.randn(2, 64, 256)
        >>> m(x).shape
        torch.Size([2, 64, 256])
    """

    def __init__(
        self,
        dim: int,
        hidden_layer_multiplier: int = 1,
        dim_out: Optional[int] = None,
        activation=nn.GELU(),
        dropout=0.0,
        *args,
        **kwargs,
    ):
        super().__init__()
        out_features = dim_out or dim
        hidden_features = hidden_layer_multiplier * dim

        self.conv_mlp = nn.Sequential(
            nn.Conv2d(dim, hidden_features, 1),
            nn.Conv2d(
                hidden_features,
                hidden_features,
                3,
                1,
                1,
                bias=True,
                groups=hidden_features,
            ),
            activation,
            nn.Conv2d(hidden_features, out_features, 1),
            nn.Dropout(dropout),
        )

        # This feedforward requires a context length which is squared, often due to 2D pooling
        self.requires_squared_context = True

    def init_weights(self, **kwargs):
        # Follow the original init, but also make it possible to initialize from the outside
        def init_module(m: nn.Module):
            if isinstance(m, nn.Conv2d):
                fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                fan_out //= m.groups
                m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
                if m.bias is not None:
                    m.bias.data.zero_()

        self.apply(init_module)

    def forward(self, x: Tensor) -> Tensor:
        # The conv layers expect NCHW, we have NLC by default
        B, L, C = x.shape
        HW = int(math.sqrt(x.shape[-2]))
        assert (
            HW**2 == L
        ), "Conv2DFeedforward requires squared context lengths"

        x = x.reshape((B, HW, HW, C)).swapdims(1, -1)

        # The actual FW, including the 2d convolutions
        x = self.conv_mlp(x)

        # back to NLC
        x = x.transpose(1, -1)
        return x.flatten(1, 2)



import math

from einops import rearrange
from torch import einsum, nn

from zeta.utils import l2norm


class LinearAttentionVision(nn.Module):
    """
    Linear Attention module that performs attention mechanism on the input feature map.

    Args:
        dim (int): The input feature map dimension.
        dim_head (int, optional): The dimension of each attention head. Defaults to 32.
        heads (int, optional): The number of attention heads. Defaults to 8.
        **kwargs: Additional keyword arguments.

    Returns:
        torch.Tensor: The output feature map after applying linear attention.

    """

    def __init__(self, dim: int, dim_head: int = 32, heads: int = 8, **kwargs):
        super().__init__()
        self.scale = dim_head**-0.5
        self.heads = heads
        inner_dim = dim_head * heads
        self.norm = nn.LayerNorm(dim)

        self.nonlin = nn.GELU()
        self.to_qkv = nn.Conv2d(dim, inner_dim * 3, 1, bias=False)

        self.to_out = nn.Sequential(
            nn.Conv2d(inner_dim, dim, 1, bias=False), nn.LayerNorm(dim)
        )

    def forward(self, fmap):
        """
        Forward pass of the LinearAttention module.

        Args:
            fmap (torch.Tensor): Input feature map tensor of shape (batch_size, channels, height, width).

        Returns:
            torch.Tensor: Output tensor after applying linear attention, of shape (batch_size, channels, height, width).
        """
        h, x, y = self.heads, *fmap.shape[-2:]
        seq_len = x * y

        fmap = self.norm(fmap)
        q, k, v = self.to_qkv(fmap).chunk(3, dim=1)
        q, k, v = map(
            lambda t: rearrange(t, "b (h c) x y -> (b h) (x y) c", h=h),
            (q, k, v),
        )

        q = q.softmax(dim=-1)
        k = k.softmax(dim=-2)

        q = q * self.scale
        v = l2norm(v)

        k, v = map(lambda t: t / math.sqrt(seq_len), (k, v))

        context = einsum("b n d, b n e -> b d e", k, v)
        out = einsum("b n d, b d e -> b n e", q, context)
        out = rearrange(out, "(b h) (x y) d -> b (h d) x y", h=h, x=x, y=y)

        out = self.nonlin(out)
        return self.to_out(out)



import torch
from torch import nn
import torch.nn.functional as F


class RMSNorm(nn.Module):
    """
    RMS  Normalization

    Args:
        dim (int): The dimension of the input.
        eps (float): The epsilon value.

    Attributes:
        scale (float): The scale value.
        gamma (nn.Parameter): The gamma parameter.

    Example:
        >>> module = RMSNorm(768)
        >>> x = torch.randn(2, 197, 768)
        >>> y = module(x)
        >>> y.shape
        torch.Size([2, 197, 768])

    """

    def __init__(self, dim, groups=1):
        super().__init__()
        self.scale = dim**-0.5
        self.gamma = nn.Parameter(torch.ones(groups, dim, 1))

    def forward(self, x):
        """Forward method implementation."""
        normed = F.normalize(x, dim=-2)
        return normed * self.scale * self.gamma



from torch.nn import Module


def set_module_requires_grad(
    module: Module,
    requires_grad: bool,
):
    """
    Set the `requires_grad` attribute of all parameters in the given module.

    Args:
        module (Module): The module whose parameters' `requires_grad` attribute needs to be set.
        requires_grad (bool): The value to set for the `requires_grad` attribute.

    Returns:
        None
    """
    for param in module.parameters():
        param.requires_grad = requires_grad


def freeze_all_layers(module):
    """
    Freezes all layers in the given module by setting their requires_grad attribute to False.

    Args:
        module (nn.Module): The module whose layers need to be frozen.
    """
    set_module_requires_grad(module, False)



import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange
import math


class DynaConv(nn.Module):
    """
    DynaConv dynamically generates convolutional kernels based on the input features.

    This layer replaces traditional convolutional layers with a dynamic mechanism,
    where convolutional kernels are generated on-the-fly by a small neural network.

    Args:
        in_channels (int): Number of channels in the input image.
        out_channels (int): Number of channels produced by the convolution.
        kernel_size (int or tuple): Size of the convolving kernel.
        stride (int or tuple, optional): Stride of the convolution. Default: 1
        padding (int or tuple, optional): Zero-padding added to both sides of the input. Default: 0
        groups (int, optional): Number of blocked connections from input channels to output channels. Default: 1
        bias (bool, optional): If True, adds a learnable bias to the output. Default: True

    Example:
        >>> dynaconv = DynaConv(in_channels=3, out_channels=64, kernel_size=3, stride=1, padding=1)
        >>> input_tensor = torch.randn(1, 3, 224, 224) # Example input batch
        >>> output = dynaconv(input_tensor)
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        groups=1,
        bias=True,
    ):
        super(DynaConv, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = (
            kernel_size
            if isinstance(kernel_size, tuple)
            else (kernel_size, kernel_size)
        )
        self.stride = stride
        self.padding = padding
        self.groups = groups

        # The small network to generate dynamic kernels. It's a simple MLP.
        self.kernel_generator = nn.Sequential(
            nn.Linear(
                in_channels * self.kernel_size[0] * self.kernel_size[1],
                out_channels,
            ),
            nn.Tanh(),
            nn.Linear(
                out_channels,
                out_channels * self.kernel_size[0] * self.kernel_size[1],
            ),
        )

        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter("bias", None)

        # Initialize parameters
        self.reset_parameters()

    def reset_parameters(self):
        gain = nn.init.calculate_gain("tanh")
        nn.init.kaiming_uniform_(self.kernel_generator[0].weight, a=gain)
        if self.bias is not None:
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(
                self.kernel_generator[0].weight
            )
            bound = 1 / math.sqrt(
                fan_in
            )  # Use math.sqrt for the scalar square root calculation
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, x):
        batch_size, _, H, W = x.shape
        x_unfold = F.unfold(
            x,
            kernel_size=self.kernel_size,
            stride=self.stride,
            padding=self.padding,
        )

        # The input to kernel_generator must match its expected input dimensions.
        # We reshape x_unfold to have dimensions [batch_size * number of patches, in_channels * kernel_size * kernel_size]
        x_unfold = rearrange(
            x_unfold,
            "b (c kh kw) l -> (b l) (c kh kw)",
            c=self.in_channels,
            kh=self.kernel_size[0],
            kw=self.kernel_size[1],
        )

        kernels = self.kernel_generator(x_unfold).view(
            batch_size,
            -1,
            self.out_channels,
            self.kernel_size[0],
            self.kernel_size[1],
        )

        # Apply the generated kernels for each patch
        output = torch.einsum(
            "blodij,blij->bod",
            kernels,
            x_unfold.view(
                batch_size,
                -1,
                self.in_channels,
                self.kernel_size[0],
                self.kernel_size[1],
            ),
        )

        # Reshape output to match the convolutional output
        output = rearrange(
            output,
            "b (h w) d -> b d h w",
            h=H // self.stride,
            w=W // self.stride,
        )

        # Add bias if necessary
        if self.bias is not None:
            output += self.bias.view(1, -1, 1, 1)

        return output


# # Example usage
# dynaconv = DynaConv(in_channels=3, out_channels=64, kernel_size=3, stride=1, padding=1)
# input_tensor = torch.randn(1, 3, 224, 224)  # Example input batch
# output = dynaconv(input_tensor)



from torch import nn, Tensor
from einops import rearrange, reduce


def video_to_text(x: Tensor, seqlen: int, dim: int, norm: bool = True):
    """
    Convert a video tensor to a text tensor.

    Args:
        x (Tensor): Input video tensor of shape (batch_size, time, channels, height, width).
        seqlen (int): Length of the output text sequence.
        dim (int): Dimension of the intermediate representation.
        norm (bool, optional): Whether to apply layer normalization. Defaults to True.

    Returns:
        Tensor: Output text tensor of shape (batch_size, seqlen, dim).

    Example::
        >>> x = torch.randn(2, 10, 3, 32, 32)
        >>> x = video_to_text(x, 100, 512)
        >>> x.shape
        torch.Size([2, 100, 512])
    """
    b, t, c, h, w = x.shape

    x = rearrange(x, "b t c h w -> b t c (h w)")
    x = reduce(x, "b t c (h w) -> b t c", "mean", h=h, w=w)
    x = nn.Linear(c, dim)(x)
    x = rearrange(x, "b t d -> b d t")
    x = nn.Linear(t, seqlen)(x)
    x = rearrange(x, "b d t -> b t d")
    return nn.LayerNorm(dim)(x)



import torch
from torch import nn


# Example usage of the IterativeCrossSelfAttention class
class PreNorm(nn.Module):
    """Prenorm

    Args:
        dim (_type_): _description_
        fn (_type_): _description_

    """

    def __init__(self, dim, fn):
        super().__init__()
        self.norm = nn.LayerNorm(dim)
        self.fn = fn

    def forward(self, x, context=None):
        """Forward pass of prenorm

        Args:
            x (_type_): _description_
        """
        return self.fn(self.norm(x), context=context)


class CrossAttention(nn.Module):
    def __init__(
        self,
        dim,
        heads: int = 8,
        dim_head: int = 64,
        dropout: float = 0.0,
        qk_norm: bool = True,
    ):
        super().__init__()
        inner_dim = dim_head * heads
        self.heads = heads
        self.scale = dim_head**-0.5

        self.attend = nn.Softmax(dim=-1)
        self.to_q = nn.Linear(dim, inner_dim, bias=False)
        self.to_kv = nn.Linear(dim, inner_dim * 2, bias=False)

        self.to_out = nn.Sequential(
            nn.Linear(inner_dim, dim), nn.Dropout(dropout)
        )

        self._qk_norm = nn.LayerNorm(dim)

    def forward(self, x, context=None):
        if context is None:
            context = x

        q = self.to_q(x)
        kv = self.to_kv(context).chunk(2, dim=-1)
        k, v = kv[0], kv[1]

        if self.qk_norm:
            q, k = self._qk_norm(q), self._qk_norm(k)

        dots = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        attn = self.attend(dots)
        out = torch.matmul(attn, v)
        out = self.to_out(out)
        return out


class IterativeCrossSelfAttention(nn.Module):
    """Iterative

    Args:
        dim (_type_): _description_
        depth (_type_): _description_
        heads (_type_): _description_
        dim_head (_type_): _description_
        dropout (float, optional): _description_. Defaults to 0.1.

    Methods:
        forward(x, context=None): _description_

    Examples:
    """

    def __init__(
        self,
        dim,
        depth,
        heads,
        dim_head,
        dropout=0.1,
    ):
        super().__init__()
        self.layers = nn.ModuleList(
            [
                PreNorm(
                    dim,
                    CrossAttention(
                        dim, heads=heads, dim_head=dim_head, dropout=dropout
                    ),
                )
                for _ in range(depth)
            ]
        )

    def forward(self, x: torch.Tensor, context: torch.Tensor = None):
        """Forward pass of IterativeCrossSelfAttention

        Args:
            x (torch.Tensor): _description_
            context (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        for layer in self.layers:
            x = layer(x, context=context) + x
        return x


# import torch

# # Example usage of the IterativeCrossSelfAttention class
# if __name__ == "__main__":
#     batch_size = 8
#     seq_len = 16  # Sequence length of the input embeddings
#     latent_seq_len = 16  # Sequence length of the latent array (could be different from input sequence length)
#     dim = 512  # Dimensionality of the input embeddings and latent array
#     heads = 8  # Number of attention heads
#     dim_head = 64  # Dimensionality of each attention head
#     depth = 6  # Number of cross-attention layers

#     # Initialize the IterativeCrossSelfAttention module
#     iter_cs_attn = IterativeCrossSelfAttention(dim, depth, heads, dim_head)

#     # Create random tensors for the input embeddings and the latent array
#     input_embeddings = torch.rand(batch_size, seq_len, dim)
#     latent_array = torch.rand(batch_size, latent_seq_len, dim)

#     # Pass the input embeddings and the latent array through the IterativeCrossSelfAttention module
#     output_embeddings = iter_cs_attn(input_embeddings, latent_array)

#     print("Output embeddings shape:", output_embeddings.shape)



import logging
import os
import warnings
import tensorflow as tf
import numexpr as ne


def disable_warnings_and_logs():
    """
    Disables various warnings and logs.
    """

    class CustomFilter(logging.Filter):
        def filter(self, record):
            unwanted_logs = [
                "Setting ds_accelerator to mps (auto detect)",
                (
                    "NOTE: Redirects are currently not supported in Windows or"
                    " MacOs."
                ),
            ]
            return not any(log in record.getMessage() for log in unwanted_logs)

    # disable warnings
    warnings.filterwarnings("ignore")

    # disable tensorflow warnings
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
    tf.get_logger().setLevel("ERROR")

    ## disable tensorflow logs
    os.getenv("TF_CPP_MIN_LOG_LEVEL", "3")

    # disable numexpr INFO logs
    ne.set_num_threads(1)
    ne.set_vml_num_threads(1)

    # disable bnb warnings and others
    logging.getLogger().setLevel(logging.ERROR)

    # add custom filter to root logger
    logger = logging.getLogger()
    f = CustomFilter()
    logger.addFilter(f)

    # disable specific loggers
    loggers = [
        "real_accelerator",
        "torch.distributed.elastic.multiprocessing.redirects",
    ]

    for logger_name in loggers:
        logger = logging.getLogger(logger_name)
        logger.setLevel(logging.CRITICAL)

    # disable all loggers
    logging.disable(logging.CRITICAL)



import torch
from torch import nn, Tensor
from zeta.nn.modules.visual_expert import VisualExpert
from zeta.nn.modules.mlp import MLP
from zeta.nn.modules.simple_mamba import MambaBlock
from zeta.structs.transformer import ViTransformerWrapper, Encoder


class MultiModalMambaBlock(nn.Module):
    """
    MultiModalMambaBlock is a PyTorch module that combines text and image embeddings using a multimodal fusion approach.

    Args:
        dim (int): The dimension of the embeddings.
        depth (int): The depth of the Mamba block.
        dropout (float): The dropout rate.
        heads (int): The number of attention heads.
        d_state (int): The dimension of the state in the Mamba block.
        image_size (int): The size of the input image.
        patch_size (int): The size of the image patches.
        encoder_dim (int): The dimension of the encoder embeddings.
        encoder_depth (int): The depth of the encoder.
        encoder_heads (int): The number of attention heads in the encoder.
        fusion_method (str): The multimodal fusion method to use. Can be one of ["mlp", "concat", "add"].

    Examples:
    x = torch.randn(1, 16, 64)
    y = torch.randn(1, 3, 64, 64)
    model = MultiModalMambaBlock(
        dim = 64,
        depth = 5,
        dropout = 0.1,
        heads = 4,
        d_state = 16,
        image_size = 64,
        patch_size = 16,
        encoder_dim = 64,
        encoder_depth = 5,
        encoder_heads = 4
    )
    out = model(x, y)
    print(out.shape)

    """

    def __init__(
        self,
        dim: int,
        depth: int,
        dropout: float,
        heads: int,
        d_state: int,
        image_size: int,
        patch_size: int,
        encoder_dim: int,
        encoder_depth: int,
        encoder_heads: int,
        fusion_method: str = "mlp",
        expansion_rate: int = 2,
        *args,
        **kwargs,
    ):
        super(MultiModalMambaBlock, self).__init__()
        self.dim = dim
        self.depth = depth
        self.dropout = dropout
        self.heads = heads
        self.d_state = d_state
        self.image_size = image_size
        self.patch_size = patch_size
        self.encoder_dim = encoder_dim
        self.encoder_depth = encoder_depth
        self.encoder_heads = encoder_heads
        self.fusion_method = fusion_method

        # Hidden dim
        self.hidden_dim = dim * expansion_rate

        # Set up the Mamba block
        self.mamba = MambaBlock(
            dim=dim, depth=depth, d_state=d_state, *args, **kwargs
        )

        # Set up the ViT encoder
        self.encoder = ViTransformerWrapper(
            image_size=image_size,
            patch_size=patch_size,
            attn_layers=Encoder(
                dim=encoder_dim,
                depth=encoder_depth,
                heads=encoder_heads,
            ),
        )

        # Setup the linear layer to project the image embeddings to the same dimension as the text embeddings
        self.linear = nn.Linear(encoder_dim, dim)

        # VisualExpert
        self.visual_expert = VisualExpert(dim, self.hidden_dim, dropout, heads)

        # MLP
        self.mlp = MLP(dim, dim, expansion_factor=4, depth=1, norm=True)

    def forward(self, text: Tensor, img: Tensor) -> Tensor:
        """
        Forward pass of the MultiModalMambaBlock module.

        Args:
            text (Tensor): The input text embeddings.
            img (Tensor): The input image.

        Returns:
            Tensor: The output embeddings after multimodal fusion.
        """
        # Encode the image, Returns the same shape as text
        encoded_img = self.encoder(img, return_embeddings=True)
        # print(f"Image shape: {encoded_img.shape} inside the MultiModalMambaBlock")
        # Project the image embeddings to the same dimension as the text embeddings
        # We need to project the 2nd dim of the image embeddings to the same dimension as the text embeddings

        # if the fusion method is mlp, use the mlp to fuse the text and image embeddings
        if self.fusion_method == "mlp":
            fusion_layer = self.mlp(encoded_img)
            fused = fusion_layer + text

        # If fusion method is concat, concatenate the text and image embeddings
        if self.fusion_method == "concat":
            fused = torch.concat([text, encoded_img], dim=1)

        if self.fusion_method == "add":
            fused = encoded_img + text

        if self.fusion_method == "visual_expert":
            concat = torch.cat([text, encoded_img], dim=1)
            fused = self.visual_expert(concat)

        return self.mamba(fused)

    def check_fusion_method(self):
        print("""[mlp] [visualexpert] [projection] [concat] [add] """)
        print(f"""Current fusion method: {self.fusion_method}""")



import torch
from einops import rearrange


def embedding_to_grid(embedding: torch.Tensor, grid_size):
    """
    Embedding to grid

    Einstein summation notation:
        'b' = batch size
        'h' = image height
        'w' = image width
        'd' = embedding dimension

    Args:
        embedding (torch.Tensor): Embedding to be reshaped
        grid_size (tuple): Grid size

    Returns:
        torch.Tensor: Reshaped embedding

    Usage:
        >>> embedding = torch.rand(1, 128, 4, 4)
        >>> embedding.shape
        torch.Size([1, 128, 4, 4])
        >>> embedding = embedding_to_grid(embedding, (8, 8))
        >>> embedding.shape
        torch.Size([1, 128, 8, 8])

    """
    h, w = grid_size
    return rearrange(embedding, "b (h w d) -> b d h w ", h=h, w=w)



from einops import rearrange


def patch_video(x, patch_size: int):
    """
    Patch a video into patches of size patch_size x patch_size x patch_size x C x H x W

    Args:
        x (torch.Tensor): Input video tensor of shape (batch_size, time, channels, height, width).
        patch_size (int): Size of the patches in each dimension.

    Returns:
        torch.Tensor: Patched video tensor of shape (batch_size, time, height, width, patch_size, patch_size, patch_size, channels).

    Example::
        >>> x = torch.randn(2, 10, 3, 32, 32)
        >>> x = patch_video(x, 4)
        >>> x.shape
        torch.Size([2, 10, 8, 8, 4, 4, 4, 3])
    """
    b, t, c, h, w = x.shape
    x = rearrange(
        x, "b t c h w -> b c t h w"
    )  # change shape to (batch_size, channels, time, height, width)
    x = rearrange(
        x,
        "b c (t p1) (h p2) (w p3) -> b t h w (p1 p2 p3) c",
        p1=patch_size,
        p2=patch_size,
        p3=patch_size,
    )
    return x



import torch
import torch.nn.functional as F


def standard_softmax(tensor):
    return F.softmax(tensor, dim=0)


# selu softmax


def selu_softmax(x):
    """
    selu_softmax works by first applying the scaled exponential linear unit
    (selu) activation function to the input tensor and then applying softmax.

    x: input tensor
    """
    # selu params
    alpha, scale = (
        1.6732632423543772848170429916717,
        1.0507009873554804934193349852946,
    )
    return F.softmax(scale * F.selu(x, alpha), dim=0)


# 2. Sparsemax


def sparsemax(x, k):
    """
    sparsemax works by first sorting the input tensor in descending order and
    then applying the following formula to the sorted tensor:

    sparsemax(z) = max(0, z - tau(z)) where tau(z) = (sum_i=1^k z_i - 1) / k

    z: input tensor
    k: number of elements to keep


    """
    original_size = x.size()
    x = x.view(-1, original_size[-1])
    dim = 1
    number_of_logits = x.size(dim)

    # Check if k is greater than the number of logits
    if k > number_of_logits:
        raise ValueError("k cannot be greater than the number of logits.")

    x = x - torch.max(x, dim=dim, keepdim=True).values
    sorted_x, _ = torch.sort(x, dim=dim, descending=True)
    cumulative_values = torch.cumsum(sorted_x, dim=dim) - 1
    range_values = torch.arange(
        start=1, end=number_of_logits + 1, device=x.device
    )
    bound = (sorted_x - cumulative_values / range_values) > 0
    rho = torch.count_nonzero(bound, dim=dim)

    # Check if k is too large and adjust it
    if k > rho.max():
        k = rho.max().item()

    tau = cumulative_values.gather(dim, rho.unsqueeze(dim) - 1)
    tau /= rho.to(dtype=torch.float32)
    return torch.max(torch.zeros_like(x), x - tau.unsqueeze(dim)).view(
        original_size
    )


# 3. Local Softmax
def local_softmax(tensor, num_chunks: int = 2):
    """
    local softmax works by splitting the input tensor into num_chunks smaller
    tensors and then applying softmax on each chunk. The results are then
    concatenated and returned.

    tensor: input tensor
    num_chunks: number of chunks to split the tensor into


    """
    # split the tensor into num chunks smaller tensor
    tensors = torch.chunk(tensor, num_chunks, dim=0)

    # apply softmax on each chunk and collect the results in a list
    results = [F.softmax(t, dim=0) for t in tensors]

    # concat results
    concated_results = torch.cat(results, dim=0)

    return concated_results


# 4. Fast Softmax


def fast_softmax(tensor):
    """
    LogSumExp trick for numerical stability

    tensor = torch.rand(10, 5)
    result = fast_softmax(tensor)
    print(result)

    """
    shiftx = tensor - torch.max(tensor)

    exps = torch.exp(shiftx)

    return exps / torch.sum(exps)


# 5. Sparse Softmax
def sparse_softmax(z, k: int = 3):
    """
    Sparsemax works by first sorting the input tensor in descending order and
    then applying the following formula to the sorted tensor:

    sparsemax(z) = max(0, z - tau(z)) where tau(z) = (sum_i=1^k z_i - 1) / k

    z: input tensor
    k: number of elements to keep

    """
    _, top_k_indices = z.topk(k, dim=0)
    omega_k = top_k_indices

    # compute sparse softmax transformation
    exp_z = torch.exp(z)
    masked_sum_exp = exp_z[omega_k].sum()
    values = torch.zeros_like(z)
    values[omega_k] = exp_z[omega_k] / masked_sum_exp

    return values


# 6. gumbelmax


def gumbelmax(x, temp=1.0, hard=False):
    """
    Gumbelmax works by adding Gumbel noise to the input tensor x and then
    applying softmax. The hard parameter controls whether the output will
    be one-hot or a probability distribution.

    x: input tensor
    temp: temperature parameter
    hard: if True, the returned tensor will be one-hot, otherwise a probability distribution

    """
    gumbels = -torch.empty_like(x).exponential_().log()
    y = x + gumbels
    y = F.softmax(y / temp, dim=-1)

    if hard:
        y_hard = torch.zeros_like(x).scatter_(
            -1, y.argmax(dim=-1, keepdim=True), 1.0
        )
        y = y_hard - y.detach() + y
    return y


# 7. Softmax with temp


def temp_softmax(x, temp=1.0):
    """
    Temp softmax works by dividing the input tensor by the temperature
    parameter and then applying softmax.

    x: input tensor
    temp: temperature parameter

    """
    return F.softmax(x / temp, dim=-1)


# 8. logit scaled softmax


def logit_scaled_softmax(x, scale=1.0):
    """
    logit scaled softmax works by multiplying the input tensor by the scale
    parameter and then applying softmax.

    x: input tensor
    scale: scale parameter
    """
    return F.softmax(x * scale, dim=-1)


# 9. norm exponential softmax


def norm_exp_softmax(x, scale=1.0):
    """
    norm exponential softmax works by applying the following formula to the
    input tensor:

    norm_exp_softmax(x) = exp(scale * x) / sum(exp(scale * x))

    x: input tensor
    scale: scale parameter
    """
    return torch.exp(scale * x) / torch.exp(scale * x).sum(dim=-1, keepdim=True)



import torch
import torch.nn as nn


class STEFunc(torch.autograd.Function):
    """
    Straight Through Estimator

    This function is used to bypass the non differentiable operations


    Args:
        input (torch.Tensor): the input tensor

    Returns:
        torch.Tensor: the output tensor

    Usage:
    >>> x = torch.randn(2, 3, requires_grad=True)
    >>> y = STEFunc.apply(x)
    >>> y.backward(torch.ones_like(x))
    >>> x.grad


    """

    @staticmethod
    def forward(ctx, input):
        """
        Forward pass of the STE function where we clamp the input between -1 and 1 and then apply the sign function

        Args:
            ctx (torch.autograd.Function): the context object
            input (torch.Tensor): the input tensor



        """
        return torch.sign(torch.clamp(input, min=-1.0, max=1.0))

    @staticmethod
    def backward(ctx, grad_output):
        """
        Backward pass of the STE function where we bypass the non differentiable operations

        """
        # Bypass the non differterable operations
        return grad_output


class STE(nn.Module):
    """
    STE Module

    This module is used to bypass the non differentiable operations

    Args:
        input (torch.Tensor): the input tensor

    Returns:
        torch.Tensor: the output tensor

    Usage:
    >>> ste = STE()
    >>> x = torch.randn(2, 3, requires_grad=True)
    >>> y = ste(x)
    >>> y.backward(torch.ones_like(x))
    >>> x.grad

    """

    def forward(self, input):
        return STEFunc.apply(input)



from torch import nn
from typing import List
import torch.nn.functional as F


class MLPProjectionFusion(nn.Module):
    def __init__(
        self,
        input_dims: List[int],
        dim: int,
        num_experts: int,
    ):
        """
        Initializes an instance of MLPProjectionFusion.

        Args:
            input_dims (List[int]): A list of input dimensions for each expert.
            dim (int): The dimension of the MLP layers.
            num_experts (int): The number of experts.

        """
        super().__init__()
        self.input_dims = input_dims
        self.dim = dim
        self.num_experts = num_experts

        # First layer MLP for each expert
        self.mlp_layers = nn.ModuleList(
            [nn.Linear(dim, dim) for dim in input_dims]
        )

        # Shared second layer of mlp2
        self.mlp2 = nn.Linear(dim, dim)

    def forward(self, *expert_inputs):
        """
        Forward pass of the MLPProjectionFusion module.

        Args:
            *expert_inputs: Variable number of expert inputs.

        Returns:
            torch.Tensor: The fused output.

        Raises:
            AssertionError: If the number of inputs does not match the number of experts.

        """
        assert (
            len(expert_inputs) == self.num_experts
        ), "Number of inputs must match number of experts"

        # Process each expert input through its mlp1 and sum the results
        expert_projections = [
            self.mlp2(F.relu(self.mlp_layers[i](input)))
            for i, input in enumerate(expert_inputs)
        ]

        # Fused output
        fused_output = sum(expert_projections)

        return fused_output
